import json
import numpy as np
from symmetr import parse, sym_res
from symmetr.tensors import Tensor

def test_sE_p1():
    opt = parse('res s E -f Mn2Au/findsym.in -p 1')
    X = sym_res(opt)
    with open('Mn2Au/sv0.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sE_p1_p2():
    opt = parse('res s E -f Mn2Au/findsym.in -p 1 -p2 2')
    X = sym_res(opt)
    with open('Mn2Au/sv_0_1.out') as f:
        res = json.load(f)
    Xr00 = Tensor.load(res['X00'])
    Xr01 = Tensor.load(res['X01'])
    Xr10 = Tensor.load(res['X10'])
    Xr11 = Tensor.load(res['X11'])

    X_zero = Tensor(0,3,2)

    assert X[0][0] - Xr00 == X_zero
    assert X[0][1] - Xr01 == X_zero
    assert X[1][0] - Xr10 == X_zero
    assert X[1][1] - Xr11 == X_zero

def test_jE():
    opt = parse('res v E -f Mn2Au/findsym.in  ')
    X = sym_res(opt)
    with open('Mn2Au/vv.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sv_1_equiv():

    opt = parse('res s E -f Mn2Au/findsym.in -p 1 -e  ')
    X = sym_res(opt)

    confs = {}
    for i in X[1].confs:
        confs[i] = {}
        for j in X[1].confs[i]:
            confs[i][j] = np.array(X[1].confs[i][j],dtype=float).squeeze()
            
    with open('Mn2Au/sv_0_equiv.out') as f:
        res = json.load(f)

    X_zero = Tensor(0,3,2)
    for i in X[1].Xs:
        for j in range(2):
            assert X[1].Xs[i][j] - Tensor.load(res['Xs'][str(i)][j]) == X_zero

    for i in confs:
        for j in confs[i]:
            resc = np.array(res['confs'][str(i)][str(j)])
            assert np.linalg.norm(confs[i][j] - resc) < 1e-10 





import json
import numpy as np
from symmetr import parse, sym_res
from symmetr.tensors import Tensor

def test_jE():
    opt = parse('res j E -f MnTe/findsym.in' )
    X = sym_res(opt)

    with open('MnTe/vv.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_svE():
    opt = parse('res s.v E -f MnTe/findsym.in' )
    X = sym_res(opt)

    with open('MnTe/sv.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,3)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_jE2():
    opt = parse('res j E -f MnTe/MnTe_inp1.in --exp 2 --syms 2,5,7' )
    X = sym_res(opt)

    with open('MnTe/cond2_rot.json') as f:
        res = json.load(f)
    Xr = Tensor.load(res)
    X_zero = Tensor(0,3,2)

    assert X - Xr == X_zero



from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import str
from builtins import range
import re

import sympy
import numpy
from .tensors import matrix, mat2ten, Tensor
from .conv_index import *


def should_rename(X,X_t):
    #looks at the element and the transformation of the element and decides if the element should be renamed

    if X - X_t == 0:
        return False
    else: 
        inds = re.findall("\[([0-9]), ([0-9])\]",str(X))
        if len(inds) > 1:
            return False
        elif len(inds) == 0:
            sys.exit('error in should_rename')
        else:
            ret = True
            inds_t = re.findall("\[([0-9]), ([0-9])\]",str(X_t))
            for ind in inds_t:
                if convert_index(inds[0][0],inds[0][1]) < convert_index(ind[0],ind[1]):
                    ret = False
            return ret


#renames the matrix so that it has the simplest form possible
#it looks for the relation between components so no information is lost
#debug is an optional parameter, if it's true, then the routine outputs lots of information
def rename(X,name,debug=False,ignore_ren_warning=False):

    ninds = len(list(set(re.findall(r'x[0-9]+',sympy.srepr(X)))))

    #v contains the symbolic variables:
    V = matrix('s',3,name)
    #Y will be the renamed matrix
    Y = matrix(0,3)

    if debug:
        print('')
        print('======= Start renaming =======')
        print('Input matrix:')
        sympy.pprint(X.mat())

        print('')
        print('independent components:')
        print(list(set(re.findall(r'x[0-9]+',sympy.srepr(X)))))
        print('')

    #a loop over all components of  the input matrix
    for i in range(3):
        for j in range(3):

            #if a components is zero we do nothing with it
            if X[i,j] == 0:
                Y[i,j] = 0
            else:

                if debug:
                    print('')
                    print('component: ', i, j, ' = ', end=' ')
                    sympy.pprint(X[i,j])

                pos = convert_index(i,j)
                if pos == 0: #we always rename the first component of the matrix (unless it's zero)
                    Y[i,j] = V.x[i,j]
                    if debug:
                        print('renaming to:', ' ', end=' ')
                        sympy.pprint(V.x[i,j])
                # if we don't have the first component we try if the component is not a linear combination
                # of the previous
                #components, if it's not we give it a new name, if it is then we set it equal to that combination
                else: 
                    #to find if the component is a linear combination of a previous components we set a linear equations system
                    #matrix of that system will be Z, last column of Z is the right hand side
                    Z = sympy.zeros(9,pos+1)
                    # al loop over all components of the Z matrix
                    for p in range(pos+1):
                        for o in range(9):

                            #we set the variable that corresponds to the column to 1 and the rest to 0
                            tmp = X[inconvert_index(p)[0],inconvert_index(p)[1]]
                            for m in range(9):
                                if m == o:
                                    tmp = tmp.subs(V.x[inconvert_index(m)[0],inconvert_index(m)[1]],1)
                                else:
                                    tmp = tmp.subs(V.x[inconvert_index(m)[0],inconvert_index(m)[1]],0)

                            Z[o,p] = tmp
                    
                    if debug:
                        print('linear equations system: (last column is the right hand side)')
                        sympy.pprint(Z)

                    #this will give a solution to the linear equation system
                    # if there is no solution it outputs none
                    #there is a sympy routine for this, but it outputs a general solution that may contain parameters
                    #solve_lin routines sets all arbitrary parameters to 0, ie if there is infinitely many solutions it outputs just one
                    solution = solve_lin(Z)
                    if solution:
                        for ii in range(len(solution)):
                            if abs(solution[ii]) < 1.e-14:
                                solution[ii] = 0
                        sol = False

                    if debug:
                        print('solution of the system:')
                        print(solution)

                    #if the component is not a linear combination of previous components we rename it
                    if not solution:
                        Y[i,j] = V.x[i,j]

                        if debug:
                            print('renaming to:', ' ', end=' ')
                            sympy.pprint(V.x[i,j])

                    #if it is we set it equal to the linear combination
                    else:
                        for r in range(len(solution)):
                            Y[i,j] = Y[i,j] + solution[r] * Y[inconvert_index(r)[0],inconvert_index(r)[1]]
                        if debug:
                            print('renaming to:', ' ', end=' ')
                            sympy.pprint(Y[i,j])
                    

    if debug:
        print('')
        print('Input matrix:')
        sympy.pprint(X.mat())
        print('')
        print('renamed matrix:')
        sympy.pprint(Y.mat())
        print('')
        print('======= end rename =======')
        print('') 

    ninds_new = len(list(set(re.findall(r'x[0-9]+',sympy.srepr(Y)))))

    if ninds != ninds_new and not ignore_ren_warning:
        print('!WARNING! Problem in renaming tensor components. Try --no-rename.') 
        print('old number of independent components:', ninds)
        print(list(set(re.findall(r'x[0-9]+',sympy.srepr(X)))))
        print('new number of independent components:', ninds_new)
        print(list(set(re.findall(r'x[0-9]+',sympy.srepr(Y)))))

    return Y


#outputs a solution to the linear equation system represented by matrix Z
#last column of Z is the right hand side
# if there is more than one solution, this chooses one solution by setting all arbitraty parameters to 0
def solve_lin(Z):

    #we need to loop over variables, this doesn't seem to be possible with sympy so this is a way around that
    b0 = sympy.Symbol('b0')
    b1 = sympy.Symbol('b1')
    b2 = sympy.Symbol('b2')
    b3 = sympy.Symbol('b3')
    b4 = sympy.Symbol('b4')
    b5 = sympy.Symbol('b5')
    b6 = sympy.Symbol('b6')
    b7 = sympy.Symbol('b7')

    def convert_b(i):
        if i == 0:
            return b0
        if i == 1:
            return b1
        if i == 2:
            return b2
        if i == 3:
            return b3
        if i == 4:
            return b4
        if i == 5:
            return b5
        if i == 6:
            return b6
        if i == 7:
            return b7

    Zr = sympy.zeros(Z.rows,Z.cols)
    for i in range(Z.rows):
        for j in range(Z.cols):
            Zr[i,j] = Z[i,j]

    #this solves  the system, not all the variabes may be needed, but luckily sympy doesn't complain about that
    solution = sympy.solve_linear_system(Zr,b0,b1,b2,b3,b4,b5,b6,b7,minimal=True,warn=True)

    #number of variables of the linear equation system
    dim = len(Z[1,:])-1
    #the list that will contain the solution
    out = []
    #if there is no solution we return None
    if not solution:
        return None
    else:
        #if there is a solution, we look at every component of the solution and set the arbitrary parameters to 0
        #the arbitrary parameters are not in the solution dict, that's how we find them
        for i in range(dim):
            x = convert_b(i)
            if x in solution:
                out_temp = solution[x]
                #we go over all variables and check if they are arbitrary
                for j in range(dim):
                    y = convert_b(j)
                    #if they are arbitrary we set them to zero
                    if y not in solution:
                        out_temp = out_temp.subs(y,0)
            else:
                out_temp = 0
            out.append(out_temp)

    return out






    



import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform import Rotation as R
from symmetr.symmetry import Symmetry
import sympy

def merge_lists(lists):
    if lists is None:
        return None
    merged = []
    for list in lists:
        if list is None:
            merged.append(None)
        else:
            merged += list
    return merged

class RotationType:
    """
    Class that is used to represent the various rotations types.
    typ:
        para: Arbitrary rotation around a given axis
        anti-para: 180 degree rotation around any axis perpendicular to the given axis
        single: a rotation given by axis and angle
        non-mag: completely arbitrary rotation
        identity: Identity matrix, i.e. zero angle rotation
    """

    def __init__(self, typ, axis=None, angle=None):
        if typ not in ['para', 'anti-para', 'single', 'non-mag','identity']:
            raise Exception('Rotation has to be one of "para", "anti-para", "single"!')
        self.typ = typ
        if self.typ not in ['non-mag','identity']:
            if axis is None:
                raise Exception('axis needs to be specified!')
            self.axis = axis / norm(axis)
            if self.typ in ['para', 'anti-para']:
                self.angle = None
                if angle is not None:
                    raise Warning('angle not need for parallel rotation and anti-parallel rotation.')
            else:
                if angle is None:
                    raise Exception('Angle needed for single rotation')
                self.angle = angle

    def __repr__(self):
        out = 'Rotation type: {}\n'.format(self.typ)
        if self.typ == 'non-mag':
            out += 'Arbitrary rotation\n'
        elif self.typ == 'identity':
            out += 'Identity (zero angle rotation)\n'
        elif self.typ in ['para', 'single']:
            out += 'Rotation axis: {}\n'.format(self.axis)
        else:
            out += 'Rotation axis: perpendicular to {}\n'.format(self.axis)

        if self.typ == 'single':
            out += 'Rotation angle: {}'.format(self.angle)
        elif self.typ == 'anti-para':
            out += 'Rotation angle: {}'.format(np.pi)
        elif self.typ == 'para':
            out += 'Rotation angle: arbitrary'
        return out

    def eq(self,other,prec):
        if self.typ != other.typ:
            return False
        if self.typ in ['identity', 'non-mag']:
            return True
        if self.typ in ['para', 'anti-para']:
            if norm(self.axis - other.axis) < prec:
                return True
            if norm(self.axis + other.axis) < prec:
                return True
            return False
        if self.typ == 'single':
            if norm(self.axis - other.axis) < prec:
                if abs(self.angle - other.angle ) < prec:
                    return True
                return False
            if norm(self.axis + other.axis) < prec:
                if abs(self.angle + other.angle - 2 * np.pi) < prec:
                    return True
                return False
            return False
        raise Exception('Wrong rot type')

class NosoSymFinder:
    """
    This class contains the functions that are used for finding the non-relativistic symmetries. They are grouped
    together in a class just so that all have access to the prec and debug parameters.

    prec: General precision parameter used throughout to determine if two floats are the same
    moment_zero: A precision parameter that is used to determine if a magnetic moment is zero. Can be used to ignore
    small moments.
    debug: Debug 0 means no debug output, 1 means standard debug output, 2 means detailed debug.
    """

    def __init__(self,prec=1e-5,moment_zero=1e-3,debug=0):
        self.prec = prec
        self.moment_zero = moment_zero
        self.debug = debug

    def get_i_chain(self,perm,chain):
        pi = perm[chain[-1]]
        if pi == chain[0]:
            return chain
        else:
            chain.append(pi)
            return self.get_i_chain(perm,chain)

    def get_i_chain_nr(self,perm,start):
        chain = []
        chain.append(start)
        finished = False
        current = start
        while not finished:
            next = perm[current]
            if next not in chain:
                chain.append(next)
                current = next
            elif next == chain[0]:
                #chain.append(next)
                finished = True
                chain_ok = True
            else:
                #chain.append(next)
                finished = True
                chain_ok = False
        return chain, chain_ok

    def get_permutation_chains(self,perm):
        """
        Get all permutation chains for given permutation.
        """
        classified = []
        chains = []
        oks = []
        for i in perm:
            if i not in classified:
                chain, ok = self.get_i_chain_nr(perm,i)
                chains.append(tuple(chain))
                oks.append(ok)
                classified += chain
        ok_all = False not in oks
        return tuple(chains), ok_all

    def get_AB_rotations(self,A, B, theta):
        """
        Finds all rotations that transform A to B by a given angle theta.
        """

        An = A / norm(A)
        Bn = B / norm(B)
        para = (norm(A - B) < self.prec)
        antipara = (norm(A + B) < self.prec)

        if para:
            return [RotationType('para', An)]

        if abs(theta) < self.prec:
            return None
        else:
            if antipara:
                if abs(theta - np.pi) > self.prec:
                    return None
                else:
                    return [RotationType('anti-para', An)]
            else:
                if np.dot(An, Bn) < np.cos(theta) - 1e-11:
                    if self.debug > 1:
                        print('angle not compatible')
                        print(np.dot(An, Bn), np.cos(theta))
                    return None

                sqrtt = 2 * (np.dot(An, Bn) - np.cos(theta)) / (1 - np.cos(theta)) / (1 + np.dot(An, Bn))
                if abs(sqrtt) < 1e-12:
                    sqrtt = 0
                rots = []
                for i,sign in enumerate([1,-1]):
                    acos = sign*np.sqrt(sqrtt)
                    if abs(acos) > 1:
                        if self.debug > 1:
                            print(acos)
                        acos = np.sign(acos)
                    alpha = np.arccos(acos)
                    nt = np.cos(alpha) * (A + B) / norm(A + B) + np.sin(alpha) * np.cross(A, B) / norm(np.cross(A, B))

                    r = R.from_rotvec(nt * theta)
                    if norm(r.apply(A) - B) > self.prec:
                        r2 = R.from_rotvec(-nt * theta)
                        if norm(r2.apply(A) - B) < self.prec:
                            if self.debug > 1:
                                print('switching axis')
                            rots.append(RotationType('single', -nt, theta))
                        else:
                            raise Exception('Rotation test failed, somethign is wrong!')
                    else:
                        rots.append(RotationType('single', nt, theta))
                rots = self.remove_repeated(rots)
                return rots


    def get_rotations_overlap(self,r1, r2):
        """
        Finds an overlap between two rotation types.
        """
        if r1 is None or r2 is None:
            return None
        if r1.typ in ['non-mag']:
            return r2
        if r2.typ in ['non-mag']:
            return r1
        if r1.typ == 'identity' or r2.typ == 'identity':
            if r1.typ != 'identity' and r2.typ == 'identity':
                r1i = r2
                r2i = r1
            else:
                r1i = r1
                r2i = r2
            if r2i.typ == 'identity':
                return r1i
            elif r2i.typ == 'para':
                return r1i
            else:
                return None
        if r1.typ != 'single' and r2.typ == 'single':
            r1i = r2
            r2i = r1
        elif (r1.typ != 'single' and r2.typ != 'single') and (r1.typ != 'para' and r2.typ == 'para'):
            r1i = r2
            r2i = r1
        else:
            r1i = r1
            r2i = r2

        if r1i.typ == 'single':

            if r2i.typ == 'single':
                if norm(r1i.axis - r2i.axis) < self.prec:
                    if abs(r1i.angle - r2i.angle) < self.prec:
                        return r1
                if norm(r1i.axis + r2i.axis) < self.prec:
                    if abs(r1i.angle + r2i.angle - 2 * np.pi) < self.prec:
                        return r1i

            elif r2i.typ == 'para':
                if norm(r1i.axis - r2i.axis) < self.prec or norm(r1i.axis + r2i.axis) < self.prec:
                    return r1i

            elif r2i.typ == 'anti-para':
                if abs(r1i.angle - np.pi) < self.prec and abs(np.dot(r1i.axis, r2i.axis)) < self.prec:
                    return r1i

            else:
                raise Exception('Wrong r2i.type')

        elif r1i.typ == 'para':

            if r2i.typ == 'para':
                if norm(r1i.axis - r2i.axis) < self.prec or norm(r1i.axis + r2i.axis) < self.prec:
                    return r1i
                else:
                    return RotationType('identity')

            elif r2i.typ == 'anti-para':
                if abs(np.dot(r1i.axis, r2i.axis)) < self.prec:
                    return RotationType('single', r1i.axis, np.pi)
            else:
                raise Exception('Wrong r2i.type')

        elif r1i.typ == 'anti-para':

            if r2i.typ == 'anti-para':
                if norm(r1i.axis - r2i.axis) < self.prec or norm(r1i.axis + r2i.axis) < self.prec:
                    return r1i
                else:
                    return RotationType('single',np.cross(r1i.axis,r2i.axis),np.pi)
            else:
                raise Exception('Wrong r2i.type')

        else:
            raise Exception('Wrong r1i.type')

        return None

    def get_rotations_overlap_multi(self, r1, r2):
        """
        Finds an overlap between two lists of rotations.
        That is it finds all overlaps between r1i and r2j where r1i are rotations from r1 and r2j rotations from r2.
        """
        if self.debug > 1:
            print('rotation_overlap_multi:')
            print('r1',r1)
            print(type(r2))
            print('r2',r2)
        overlaps = []
        if r1 is None or r2 is None:
            return None
        for ri in r1:
            for rj in r2:
                r = self.get_rotations_overlap(ri,rj)
                if r is not None:
                    overlaps.append(r)
        if len(overlaps) > 0:
            overlaps = self.remove_repeated(overlaps)
            return overlaps
        else:
            return None

    def remove_repeated(self,rots):
        if rots is None:
            return None
        if len(rots) == 0:
            return rots
        rots_u = []
        rots_u.append(rots.pop(0))
        for rot in rots:
            is_in = False
            for rot2 in rots_u:
                if rot.eq(rot2,self.prec):
                    is_in = True
            if not is_in:
                rots_u.append(rot)
        return rots_u

    def merge_rotations(self,rots):
        """
        Takes a list of rotations, where each element is a list of rotations and find a common overlap.
        """
        if self.debug > 1:
            print('starting merge')
            print(rots)

        #This part seems wrong and unnecessary
        #rots_all = merge_lists(rots)
        #if any(x is None for x in rots_all):
        #    if self.debug > 1:
        #        print('finished merge')
        #    return None
        rots_merged = None
        if len(rots) == 1:
            rots_merged = rots[0]
        else:
            if self.debug > 1:
                print('starting get_rotations_overlap_multi')
            ro = self.get_rotations_overlap_multi(rots[0],rots[1])
            if self.debug > 1:
                print('finished get_rotations_overlap_multi')
            if ro is None:
                if self.debug > 1:
                    print('finished merge')
                return None
            rots_merged = ro
            for irot in range(2,len(rots)):
                    if self.debug > 1:
                        print('starting get_rotations_overlap_multi')
                    ro = self.get_rotations_overlap_multi(rots_merged,rots[irot])
                    if self.debug > 1:
                        print('finished get_rotations_overlap_multi')
                    if ro is None:
                        if self.debug > 1:
                            print('finished merge')
                        return None
                    else:
                        rots_merged = ro
        if self.debug > 1:
            print('finished merge')
        return rots_merged


    def get_chain_rotations(self, mchain, hasT=False):
        """
        Finds spin rotations for a given chain of magnetic moments.
        """

        if self.debug > 1:
            print('Magnetic chain: ', mchain)

        rotations = []
        para = False
        if not hasT:
            col_dif = max([norm(m - mchain[0]) for m in mchain])
            if col_dif < self.prec:
                para = True
        elif len(mchain) % 2 == 0:
            mchain_f = [2 * (i % 2 - 1 / 2) * m for i, m in enumerate(mchain)]
            col_dif = max([norm(m - mchain_f[0]) for m in mchain_f])
            if col_dif < self.prec:
                para = True

        if para:
            rotations.append(RotationType('para', mchain[0]))
            return rotations

        p = len(mchain)
        if not hasT or len(mchain) % 2 == 0:
            thetas = [2 * np.pi * i / p for i in range(1, int(np.floor(p / 2)) + 1)]
        else:
            thetas = [np.pi * (2 * i + 1) / p for i in range(0, int(np.ceil(p / 2)))]
        for theta in thetas:

            if self.debug > 1:
                print('theta: ', theta)

            rotations_theta = []
            for im in range(len(mchain)):
                im2 = im + 1
                if im2 > len(mchain) - 1:
                    im2 = 0
                if not hasT:
                    f = 1
                else:
                    f = -1
                rot = self.get_AB_rotations(mchain[im], f * mchain[im2], theta)
                if self.debug > 1:
                    print('A: ', mchain[im], 'B: ', mchain[im2])
                    print(rot)
                rotations_theta.append(rot)

            rotations_theta_merged = self.merge_rotations(rotations_theta)
            if rotations_theta_merged is not None:
                rotations.append(rotations_theta_merged)
        rotations_all = merge_lists(rotations)
        return rotations_all

    def get_mchain(self,chain,moments):
        mchain = [moments[i-1] for i in chain]
        mchain_norms = [norm(m) for m in mchain]
        if max(mchain_norms) < self.moment_zero:
            return 0
        if min(mchain_norms) < max(mchain_norms) - 2 * self.moment_zero:
            #raise Exception('Chain contains moments with different magnitudes!')
            return None
        return [m/mchain_norms[i] for i,m in enumerate(mchain)]

    def get_permutations_rotations(self, chains, mags, hasT=False):
        """
        Finds spin rotations for a give permutation, which is represented by the chains.
        """
        # chains = get_permutation_chains(perm)
        rots_chains = []
        for chain in chains:
            mchain = self.get_mchain(chain, mags)
            if self.debug > 0:
                print('chain: ', chain)
                print('mchain: ', mchain)
            if mchain == 0:
                rots_chains.append([RotationType('non-mag')])
            elif mchain is None:
                return None
            else:
                rots = self.get_chain_rotations(mchain, hasT=hasT)
                if self.debug > 0:
                    print(rots)
                #if len(rots) > 1:
                #    raise Exception("More than one rotation per chain, this is not expected behavior, something is wrong.")
                if len(rots) == 0:
                    return None
                else:
                    rots_chains.append(rots)
        rot_merged = self.merge_rotations(rots_chains)

        return rot_merged

    def find_noso_syms(self, syms, mags, n_discretize=4):
        """
        Determinese the non-relativistic symmetry operations of a magnetic system.

        If the system contains continuous spin-rotation symmetry, we discretize it. This is not a perfect solution
        but should be fine in most cases. This only concerns collinear systems.

        Args:
            syms ([class Symmetry]): List of the non-magnetic symmetry operations of the system. Must be in
                some cartesian basis
            mags ([np.array]): List of the magnetic moments of the system. Must be in some cartesian basis.

        returns:
            noso_syms: List of the non-relativistic symmetries.
        """

        nonmag = False
        collinear = False
        if max([norm(m) for m in mags]) < self.moment_zero:
            nonmag = True
        else:
            non_zero_mags = []
            for m in mags:
                if norm(m) > self.moment_zero:
                    non_zero_mags.append(m)
            if self.debug > 0:
                print(non_zero_mags)
            #dots = [abs(np.dot(non_zero_mags[0], m) / norm(non_zero_mags[0]) / norm(m)) for m in non_zero_mags]
            #if min(dots) > 1 - self.prec:
            #    collinear = True
            non_zero_mags_norm = [m/norm(m) for m in non_zero_mags]
            col_dif = max([min(norm(m + non_zero_mags_norm[0]), norm(m - non_zero_mags_norm[0])) for m in non_zero_mags_norm])
            if col_dif < self.prec:
                collinear = True

        syms_noso = []
        if nonmag:
            raise Exception('Not implemented')
        elif collinear:
            col_axis = non_zero_mags[0] / norm(non_zero_mags[0])
            para_rot = RotationType('para', axis=non_zero_mags[0] / norm(non_zero_mags[0]))
            triv_perm = {i: i for i in syms[0].permutations}
            sym_I = Symmetry(sympy.diag(1, 1, 1), False, permutations=triv_perm)
            if self.debug > 0:
                print('Collinear system, axis: {}'.format(col_axis))
            if n_discretize is None:
                sym_I.Rs = para_rot
                syms_noso.append(sym_I)
            else:
                for theta in np.linspace(0, np.pi, n_discretize, endpoint=True):
                    sym_noso = sym_I.copy()
                    r = R.from_rotvec(para_rot.axis * theta)
                    sym_noso.Rs = sympy.Matrix(r.as_matrix())
                    syms_noso.append(sym_noso)

        if self.debug > 0:
            print('Determining spin rotations')
            print(' ')
        chains_all = []
        all_ok = True
        for sym in syms:
            chains, ok =  self.get_permutation_chains(sym.permutations)
            if ok:
                chains_all.append(chains)
            else:
                all_ok = False
                if self.debug > 0:
                    print('WARNING Symmetry that is not compatible with the magnetic unit cell is present!'
                                 'Skipping this symmetry!')
                    print(sym)
                    print(chains)
        if not all_ok:
            print('WARNING Symmetry that is not compatible with the magnetic unit cell is present!')

        chains_u = list(set(chains_all))
        rots = []
        rotsT = []
        for i, chains in enumerate(chains_u):
            if self.debug > 0:
                print('')
                print('taking permutation chains {}'.format(chains))
            # print(chains)
            if self.debug > 0:
                print('')
                print('spin-rotation without T')
            rot = self.get_permutations_rotations(chains, mags, False)
            rot = self.remove_repeated(rot)
            if self.debug > 0:
                print('Compatible rotation: ',rot)
                print('spin-rotation with T')
            rotT = self.get_permutations_rotations(chains, mags, True)
            rotT = self.remove_repeated(rotT)
            rots.append(rot)
            rotsT.append(rotT)
            if self.debug > 0:
                print('Compatible rotation: ', rotT)

        for i, sym in enumerate(syms):
            chains,ok = self.get_permutation_chains(sym.permutations)
            if not ok:
                continue
            index = chains_u.index(chains)
            if self.debug > 0:
                print('')
                print('Taking symmetry {}'.format(i))
                print(sym)
                print('')
            if not sym.has_T:
                rot = rots[index]
            else:
                rot = rotsT[index]
            if rot is None:
                if self.debug > 0:
                    print('No compatible spin rotation')
                continue
            else:
                if self.debug > 0:
                    print('Compatible spin rotation:')
                    print(rot)
            if collinear:
                if len(rot) > 1:
                    raise Exception('More than one rotation for collinear, this should not happen.')
                rot = rot[0]
                if rot.typ == 'single' or rot.typ == 'non-mag' or rot.typ == 'identity':
                    raise Exception('Single, non-mag or identity rotation present for collinear system, something is wrong')
                else:
                    sym_noso = sym.copy()
                    if sym.has_T:
                        f = sympy.diag(-1,-1,-1)
                    else:
                        f = sympy.diag(1,1,1)
                    if rot.typ == 'para':
                        sym_noso.Rs = f * sympy.diag(1, 1, 1)
                    else:
                        rot_axis = None
                        for rot_axis_c in [np.array([1, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1]),
                                           np.array([1, 1, 0])]:
                            if abs(np.dot(rot.axis, rot_axis_c)) < self.prec:
                                rot_axis = rot_axis_c
                                break
                        if rot_axis is None:
                            rot_axis = np.cross(rot.axis, [0, 0, 1])
                            rot_axis = rot_axis / norm(rot_axis)
                        sym_noso.Rs = f * sympy.Matrix(R.from_rotvec(rot_axis * np.pi).as_matrix())
                    syms_noso.append(sym_noso)
            else:
                if len(rot) > 1:
                    raise Exception('More than one rotation, this is not implemented (and may suggest somethings wrong).')
                rot = rot[0]
                if rot.typ not in ['single','identity']:
                    raise Exception('Non-collinear system and non-single rotation present, someting is wrong!')

                if rot.typ == 'single':
                    r = R.from_rotvec(rot.axis * rot.angle).as_matrix()
                else:
                    r = np.diag([1,1,1])
                sym_noso = sym.copy()
                if not sym.has_T:
                    sym_noso.Rs = sympy.Matrix(r)
                else:
                    sym_noso.Rs = sympy.Matrix(np.matmul(np.diag([-1, -1, -1]), r))
                syms_noso.append(sym_noso)

        return syms_noso




import json
from symmetr import parse, sym_res
from symmetr.tensors import Tensor

def test_sE_exp1():
    opt = parse('res s E --exp 1 -f NiMnSb/NiMnSb.in_nonmag ')
    X = sym_res(opt)
    with open('NiMnSb/sE_exp_1.out') as f:
        res = json.load(f)
    Xr = Tensor.load(res['X'])
    X_zero = Tensor(0,3,2)

    assert X - Xr == X_zero

def test_sE_exp2():
    opt = parse('res s E --exp 2 -f NiMnSb/NiMnSb.in_nonmag ')
    X = sym_res(opt)
    with open('NiMnSb/sE_exp_2.out') as f:
        res = json.load(f)
    Xr = Tensor.load(res['X'])
    X_zero = Tensor(0,3,2)

    assert X - Xr == X_zero



import json
import numpy as np
from symmetr import parse, sym_res
from symmetr.tensors import Tensor
from symmetr.symT import get_syms_noso

def test_svE_IrMn3():
    opt = parse('res s.v E -f IrMn3/findsym.in --noso' )
    X = sym_res(opt)

    with open('IrMn3/svE_noso.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,3)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sE_Mn2Au():
    opt = parse('res s E -f Mn2Au/findsym.in --noso' )
    X = sym_res(opt)

    with open('Mn2Au/sE_noso.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_svE_NiF2():
    opt = parse('res s.v E -f NiF2/findsym.in --noso' )
    X = sym_res(opt)

    with open('NiF2/svE_noso.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,3)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sE_080():
    opt = parse('res s E -f 0.80/findsym.in --noso' )
    X = sym_res(opt)

    with open('0.80/sE_noso.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sE_080_p1():
    opt = parse('res s E -f 0.80/findsym.in --noso -p 1' )
    X = sym_res(opt)

    with open('0.80/sE_1_noso.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_nsyms_IrMn3():
    syms = get_syms_noso(parse('res j E -f IrMn3/findsym.in'))
    assert len(syms) == 96

def test_nsyms_NiF2():
    syms = get_syms_noso(parse('res j E -f NiF2/findsym.in'))
    assert len(syms) == 36

def test_nsyms_NiF2():
    syms = get_syms_noso(parse('res j E -f 0.80/findsym.in'))
    assert len(syms) == 32



import numpy as np


def get_nonzero_min_abs(A, i1, i2, j1, j2):
    """
    return idx = argmin_{i, j} abs(A[i, j]) s.t. (i1 <= i < i2 and j1 <= j < j2 and A[i, j] != 0)
    if failed, return (None, None)
    """
    idx = (None, None)
    valmin = None

    for i in range(i1, i2):
        for j in range(j1, j2):
            if A[i, j] == 0:
                continue
            if (valmin is None) or (np.abs(A[i, j]) < valmin):
                idx = (i, j)
                valmin = np.abs(A[i, j])
    return idx


def get_nonzero_min_abs_full(A, s):
    """
    return idx = argmin_{i, j} abs(A[i, j]) s.t. (i >= s and j >= s and A[i, j] != 0)
    if failed, return (None, None)
    """
    return get_nonzero_min_abs(A, s, A.shape[0], s, A.shape[1])


def get_nonzero_min_abs_row(A, i1, j1):
    """
    return idx = argmin_{i, j} abs(A[i, j]) s.t. (i >= i1 and j == j1 and A[i, j] != 0)
    if failed, return (None, None)
    """
    return get_nonzero_min_abs(A, i1, A.shape[0], j1, j1 + 1)


def get_nonzero_min_abs_column(A, i1, j1):
    """
    return idx = argmin_{i, j} abs(A[i, j]) s.t. (i == i1 and j >= j1 and A[i, j] != 0)
    if failed, return (None, None)
    """
    return get_nonzero_min_abs(A, i1, i1 + 1, j1, A.shape[1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Contains functions for obtaining list of symmetry operations and transformation matrix
based on user input.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from builtins import range
from past.utils import old_div
from copy import deepcopy
from itertools import product

from . import fslib
import sympy
from sympy import sympify as spf
from .groups import group_sym
from .noso import noso_syms
from .noso_new import NosoSymFinder
from .symmetry import findsym2sym, matsym2sym, Symmetry, create_I, create_P, create_T
from fractions import Fraction
import numpy as np
from mpmath import lu_solve
from numpy.linalg import norm

from .hsnf import smith_normal_form

def make_rational(mat):
    """Converts sympy matrix to a rational number form.

    This is useful since sympy can work exactly with rational numbers so there are no floating point accuracy issues.
    """

    ncols = mat.cols
    nrows = mat.rows
    mat_r = sympy.zeros(mat.rows,mat.cols)

    for i in range(nrows):
        for j in range(ncols):
            if type(mat[i,j]) == str and '\\' in mat[i,j]:
                mat_r[i,j] = sympy.sympify(Fraction(mat[i,j]))
            else:
                mat_r[i,j] = sympy.sympify(Fraction(float(mat[i,j])))

    return mat_r

def create_Tm(vec_a,vec_b,vec_c):

    T = sympy.zeros(3)

    T[0,0] = vec_a[0]
    T[1,0] = vec_a[1]
    T[2,0] = vec_a[2]

    T[0,1] = vec_b[0]
    T[1,1] = vec_b[1]
    T[2,1] = vec_b[2]

    T[0,2] = vec_c[0]
    T[1,2] = vec_c[1]
    T[2,2] = vec_c[2]

    T = make_rational(T)

    return T

def create_Ti(fin):

    T_i = sympy.zeros(3)

    inp_type = int(fin[2])

    if inp_type == 1:

        vec_1 = fin[3].split()
        vec_2 = fin[4].split()
        vec_3 = fin[5].split()

        T_i[0,0] =  spf(vec_1[0])
        T_i[1,0] =  spf(vec_1[1])
        T_i[2,0] =  spf(vec_1[2])

        T_i[0,1] =  spf(vec_2[0])
        T_i[1,1] =  spf(vec_2[1])
        T_i[2,1] =  spf(vec_2[2])

        T_i[0,2] =  spf(vec_3[0])
        T_i[1,2] =  spf(vec_3[1])
        T_i[2,2] =  spf(vec_3[2])

    if inp_type == 2:

        a,b,c,al,bet,gam = fin[3].split()
        al = al+'*2*pi/360'
        bet = bet+'*2*pi/360'
        gam = gam+'*2*pi/360'
        gam2 = 'acos((cos({gam})-cos({bet})*cos({al}))/(sin({al})*sin({al})))'.format(gam=gam,al=al,bet=bet)        

        
        T_i[0,0] =  spf('{a}*sin({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
        T_i[1,0] =  spf('{a}*cos({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
        T_i[2,0] =  spf('{a}*cos({bet})'.format(a=a,gam2=gam2,bet=bet))

        T_i[0,1] =  spf(0)
        T_i[1,1] =  spf('{b}*sin({al})'.format(b=b,al=al))
        T_i[2,1] =  spf('{b}*cos({al})'.format(b=b,al=al))

        T_i[0,2] =  spf(0)
        T_i[1,2] =  spf(0)
        T_i[2,2] =  spf(c)
    
    return T_i

def is_hex(lines):
    for i,line in enumerate(lines):
        if 'Values of a,b,c,alpha,beta,gamma:' in line:
            pos = i
    angles = lines[pos+1].split()[3:6]
    hexag= False
    for angle in angles:
        if int(round(float(angle))) != 90:
            if int(round(float(angle))) == 120:
                hexag = True
            else:
                for i,line in enumerate(lines):
                    if 'Space Group' in line:
                        sg = line.split()[-1]
                if sg == 'P1' or sg == 'P-1':
                    hexag = False
                    break
                raise Exception('one of the angles in findsym output is neither 90 nor 120.')
    return hexag

def get_syms(opt):
    if opt['inp']:
        #runs findsym and reads the output
        lines = fslib.run_fs(opt['inp'])
        syms = fslib.r_sym(lines,num_prec = opt['pos_prec'])

    if opt['group']:
        atom = -1
        _,syms=group_sym(opt['group'],debug=False)

    #this selects some of the symmetries 
    if opt['syms_sel'] != -1:

        syms_new = []
        for i in range(len(syms)):
            if i+1 in opt['syms_sel']:
                syms_new.append(syms[i])

        syms = syms_new

    syms_g = []
    for sym in syms:
        syms_g.append(findsym2sym(sym))
    syms = syms_g

    if opt['simplify_syms']:
        idxs,syms = simplify_symmetry_operations(syms,
                                                 generators=opt['generators'],
                                                 remove_P=opt['remove_P'],
                                                 remove_T=opt['remove_T']
                                                 )
        idxs1 = [i+1 for i in idxs]
        print('Using symmetry operations: {}'.format(idxs1))

    return syms

def get_T(opt,nonmag=False):
    """
    Returns transformation matrix from the conventional coordinate system (used by findsym)
    to the user selected basis.

    There are two different behaviors:
        nonmag=False: conventional coordinate system for the magnetic system
        nonmag=True: conventional coordinate system for the nonmagnetic system
    """
    if opt['inp'] is not None:
        #runs findsym and reads the output
        fin = fslib.read_fs_inp(opt['inp'],clean=False)
        lines = fslib.run_fs(opt['inp'])
        if nonmag:
            lines_nm = fslib.run_fs_nonmag(opt['inp'])

        #reads the input
        #vec_a,b,c are needed to know the basis transformation
        #syms contain the symmetries in the form that is needed by symmetr
        if not nonmag:
            [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
        else:
            [vec_a,vec_b,vec_c] = fslib.r_basis(lines_nm)

        #construct the transformation matrix from the findsym basis to the selected basis
        if 'abc' == opt['basis']:

            T = sympy.Matrix(sympy.Identity(3))

        if 'i' == opt['basis']:

            T = create_Tm(vec_a,vec_b,vec_c)

        if 'cart' == opt['basis']:

            T_m = create_Tm(vec_a,vec_b,vec_c)

            T_i = create_Ti(fin)

            T = T_i*T_m

        if 'custom' == opt['basis']:
            
            #T_i is a transformation matrix from the input coordinate system to the cartesian one
            T_i = create_Ti(fin)
                    
            for i in range(len(fin)):
                if 'axes:' in fin[i]:
                    loc = i
            
            vec_1 = fin[loc+1].split()
            vec_2 = fin[loc+2].split()
            vec_3 = fin[loc+3].split()

            #T_c is the transformation matrix from the used-defined basis to the cartesian one
            T_c = sympy.zeros(3)

            T_c[0,0] =  spf(vec_1[0])
            T_c[1,0] =  spf(vec_1[1])
            T_c[2,0] =  spf(vec_1[2])

            T_c[0,1] =  spf(vec_2[0])
            T_c[1,1] =  spf(vec_2[1])
            T_c[2,1] =  spf(vec_2[2])

            T_c[0,2] =  spf(vec_3[0])
            T_c[1,2] =  spf(vec_3[1])
            T_c[2,2] =  spf(vec_3[2])

            normalize = True
            if normalize == True:
                norm = sympy.sqrt(T_c[0,0]**2 + T_c[1,0]**2 + T_c[2,0]**2)
                T_c[0,0] = old_div(T_c[0,0], norm)
                T_c[1,0] = old_div(T_c[1,0], norm)
                T_c[2,0] = old_div(T_c[2,0], norm)

                norm = sympy.sqrt(T_c[0,1]**2 + T_c[1,1]**2 + T_c[2,1]**2)
                T_c[0,1] = old_div(T_c[0,1], norm)
                T_c[1,1] = old_div(T_c[1,1], norm)
                T_c[2,1] = old_div(T_c[2,1], norm)

                norm = sympy.sqrt(T_c[0,2]**2 + T_c[1,2]**2 + T_c[2,2]**2)
                T_c[0,2] = old_div(T_c[0,2], norm)
                T_c[1,2] = old_div(T_c[1,2], norm)
                T_c[2,2] = old_div(T_c[2,2], norm)

            T_m = create_Tm(vec_a,vec_b,vec_c)

            T = T_c.inv()*T_i*T_m

        if 'abc_c' == opt['basis']:

            T = sympy.zeros(3)

            if not nonmag:
                abc = fslib.r_abc(lines)
            else:
                abc = fslib.r_abc(lines_nm)

            a = abc[0]
            b = abc[1]
            c = abc[2]
            al = abc[3]
            bet = abc[4]
            gam = abc[5]

            al = al+'*2*pi/360'
            bet = bet+'*2*pi/360'
            gam = gam+'*2*pi/360'
            gam2 = 'acos((cos({gam})-cos({bet})*cos({al}))/(sin({al})*sin({al})))'.format(gam=gam,al=al,bet=bet)        
            
            T[0,0] =  spf('{a}*sin({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
            T[1,0] =  spf('{a}*cos({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
            T[2,0] =  spf('{a}*cos({bet})'.format(a=a,gam2=gam2,bet=bet))

            T[0,1] =  spf(0)
            T[1,1] =  spf('{b}*sin({al})'.format(b=b,al=al))
            T[2,1] =  spf('{b}*cos({al})'.format(b=b,al=al))

            T[0,2] =  spf(0)
            T[1,2] =  spf(0)
            T[2,2] =  spf(c)

    if opt['group'] is not None:
        atom = -1
        hex_group,_=group_sym(opt['group'],debug=False)

        if 'i' == opt['basis'] or 'abc' == opt['basis']:
            print('Using the conventional coordinate system!')
            T = sympy.Matrix(sympy.Identity(3))

        if 'cart' == opt['basis']:

            print('Using a cartesian coordinate system')
            if hex_group:
                T = sympy.zeros(3)
                T[0,0] = 1
                T[0,1] = sympy.sympify(Fraction(-0.5))
                T[0,2] = 0
                T[1,0] = 0 
                T[1,1] = old_div(sympy.sqrt(3),2)
                T[1,2] = 0
                T[2,0] = 0
                T[2,1] = 0
                T[2,2] = 1
            else:
                T = sympy.Matrix(sympy.Identity(3))

    return T

def get_syms_nonmag(opt):
    lines = fslib.run_fs_nonmag(opt['inp'])
    syms = fslib.r_sym(lines)
    
    syms_g = []
    for sym in syms:
        syms_g.append(findsym2sym(sym))
    syms = syms_g

    if opt['syms_sel'] != -1:

        syms_new = []
        for i in range(len(syms)):
            if i+1 in opt['syms_sel']:
                syms_new.append(syms[i])

        syms = syms_new

    if opt['simplify_syms']:
        idxs,syms = simplify_symmetry_operations(syms,
                                                 generators=opt['generators'],
                                                 remove_P=opt['remove_P'],
                                                 remove_T=opt['remove_T']
                                                 )
        idxs1 = [i+1 for i in idxs]
        print('Using symmetry operations: {}'.format(idxs1))

    return syms

def check_nonmag_syms(syms, prec=1e-5):
    n_syms = len(syms)
    all_good = True

    syms_fix = []
    for sym in syms:
        syms_fix.append(sym.copy())

    if n_syms % 2 != 0:
        print('odd number of symmetries!!!')
        return False, None

    sym0 = syms[0]
    if sym0.has_T:
        print('first symmetry has time-reversal, that should not happen')
        return False, None

    found_shift = False
    for i, sym in enumerate(syms):
        if ((sym0.R - sym.R).norm() < prec and
                sym0.permutations == sym.permutations and
                sym.has_T):
            found_shift = True
            shift = i
            break
    if not found_shift:
        print('Time-reversal variant of the first symmetry not found!')
        return False, None

    for i, sym in enumerate(syms):
        if i + shift == len(syms):
            break
        if sym.has_T:
            continue
        sym_op = syms[i + shift]
        if not sym_op.has_T:
            print(i,shift)
            print(sym)
            print(sym_op)
            print("Shifted symmetry does not have time-reversal, don't know what to do!")
            return False, None
        if (sym.R - sym_op.R).norm() > prec or sym.permutations != sym_op.permutations:
            print('Findsym problem with non-relativistic symmetry {}, fixing.'.format(i + shift))
            all_good = False
            syms_fix[i + shift] = sym.copy()
            syms_fix[i + shift].has_T = True
            syms_fix[i + shift].Rs = -sym.Rs
    return all_good, syms_fix

def get_syms_noso(opt):
    fin_c = fslib.read_fs_inp(opt['inp'])
    mags = fslib.r_mag_fin(fin_c)
    lines = fslib.run_fs(opt['inp'])
    lines_nm = fslib.run_fs_nonmag(opt['inp'])
    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    [vec_a_nm, vec_b_nm, vec_c_nm] = fslib.r_basis(lines_nm)

    Tm = create_Tm(vec_a,vec_b,vec_c)
    Tnm = create_Tm(vec_a_nm,vec_b_nm,vec_c_nm)
    Ti = create_Ti(fin_c)
    for mag in mags:
        for i in range(3):
            mag[i] = mag[i] / Ti[:,i].norm()
    mags_T = convert_vecs(mags,Ti)
    mags_T_np = []
    for mag in mags_T:
        mags_T_np.append(np.array(mag, dtype=float).squeeze())

    syms_nm = get_syms_nonmag(opt)

    for sym in syms_nm:
        sym.convert(Ti*Tnm,in_place=True)

    if len(mags) > len(syms_nm[0].permutations):
        perms = get_full_permutations(lines_nm,debug=False)
        syms_nm_full = []
        for i, sym in enumerate(syms_nm):
            for j in range(len(perms[i])):
                sym_n = sym.copy()
                sym_n.permutations = perms[i][j]
                syms_nm_full.append(sym_n)
        syms_nm = syms_nm_full

    #there is a bug in findsym, where sometimes in a non-magnetic case, there is a wrong symmetry
    #There should be a symmetry with time-reversal for every symmetry without time-reversal, but
    #sometimes this is not the case
    #This can fix it, lathough it's more of a hack and ideally we should transition to newer version
    #of findsym, where this shouldn't happen.
    #This has to be run after the get_full_permutations!
    syms_good, syms_nm_fix = check_nonmag_syms(syms_nm)
    if not syms_good:
        if syms_nm_fix is None:
            raise Exception('Problem with non-magnetic symmetries.')
        else:
            print('fixing')
            syms_nm = syms_nm_fix

    symfinder = NosoSymFinder(prec=opt['noso_prec'],moment_zero=opt['noso_moment_zero'],debug=opt['noso_debug'])
    syms_noso = symfinder.find_noso_syms(syms_nm,mags_T_np)

    #we convert to the magnetic findsym basis, since this is what's used in the code elsewhere
    #this could be avoided though since we will just transform straight back in most situations
    for sym in syms_noso:
        sym.convert(Tm.inv()*Ti.inv(),in_place=True)

    if opt['syms_sel_noso'] != -1:
        syms_sel_noso = opt['syms_sel_noso'].split(',')
        syms_sel_noso2 = []
        for i in range(len(syms_sel_noso)):
            if '-' in syms_sel_noso[i]:
                s = syms_sel_noso[i].split('-')
                syms_sel_noso2 += list(range(int(s[0]),int(s[1])+1))
            else:
                syms_sel_noso2.append(int(syms_sel_noso[i]))

        syms_noso_new = []
        for i in range(len(syms_noso)):
            if i+1 in syms_sel_noso2:
                syms_noso_new.append(syms_noso[i])

        syms_noso = syms_noso_new

    return syms_noso

def get_syms_noso_old(opt):

    fin_c = fslib.read_fs_inp(opt['inp'])
    mags = fslib.r_mag_fin(fin_c)
    lines = fslib.run_fs(opt['inp'])
    lines_nm = fslib.run_fs_nonmag(opt['inp'])
    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    [vec_a_nm,vec_b_nm,vec_c_nm] = fslib.r_basis(lines_nm)

    Tm = create_Tm(vec_a,vec_b,vec_c)
    Tnm = create_Tm(vec_a_nm,vec_b_nm,vec_c_nm)
    mags_T = convert_vecs(mags,Tm.inv())
    syms_nm = get_syms_nonmag(opt)
    for sym in syms_nm:
        sym.convert(Tm.inv()*Tnm,in_place=True)

    perms = get_full_permutations(lines_nm)
    for i,sym in enumerate(syms_nm):
        sym.permutations = perms[i]

    hexag = is_hex(lines)
    syms_noso = noso_syms(syms_nm,mags_T,hexag,debug=opt['debug_noso'])

    if opt['syms_sel_noso'] != -1:
        syms_sel_noso = opt['syms_sel_noso'].split(',')
        syms_sel_noso2 = []
        for i in range(len(syms_sel_noso)):
            if '-' in syms_sel_noso[i]:
                s = syms_sel_noso[i].split('-')
                syms_sel_noso2 += list(range(int(s[0]),int(s[1])+1))
            else:
                syms_sel_noso2.append(int(syms_sel_noso[i]))

        syms_noso_new = []
        for i in range(len(syms_noso)):
            if i+1 in syms_sel_noso2:
                syms_noso_new.append(syms_noso[i])

        syms_noso = syms_noso_new

    return syms_noso

def get_mags(inp):

    fin_c = fslib.read_fs_inp(inp)
    mags = fslib.r_mag_fin(fin_c)

    return mags

def convert_vec(vec,T):
    vec_T = T*vec.T 
    vec_T = vec_T.T
    return vec_T

def convert_vecs(mags,T):
    """Converts a list of vectors by transform matrix T.
    """

    mags_T = []
    for i,mag in enumerate(mags): 
        mag_T = convert_vec(mag,T)
        mags_T.append(mag_T)

    return mags_T

def get_Tm(inp,nonmag=False):

    if not nonmag:
        lines = fslib.run_fs(inp)
    else:
        lines = fslib.run_fs_nonmag(inp)

    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    Tm = create_Tm(vec_a,vec_b,vec_c)

    return Tm

def get_Tm_fin(inp_list,nonmag=False):
    if not nonmag:
        lines = fslib.run_fs_fin(inp_list)
    else:
        lines = fslib.run_fs_fin(fslib.make_fsinp_nonmag(inp_list))

    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    Tm = create_Tm(vec_a,vec_b,vec_c)

    return Tm

def get_metric(opt,debug=False,nonmag=False):

    if opt['group'] is not None:
        #the metric tensor is trivial in the cartesian coordinate system
        G = sympy.Matrix([[1,0,0],[0,1,0],[0,0,1]])
        opt2 = deepcopy(opt)
        T = get_T(opt2)
        G = T.inv() * G * T.inv().T

    else:
        if debug:
            print('')
            print('Get_metric debug output')
            print('================================')
        #read the basis vectors used by findsym
        #they are given in the basis used in the findsym input
        if not nonmag:
            lines = fslib.run_fs(opt['inp'])
        else:
            lines = fslib.run_fs_nonmag(opt['inp'])
        [vec_a,vec_b,vec_c] = fslib.r_basis(lines)

        vec_a = sympy.Matrix(vec_a)
        vec_b = sympy.Matrix(vec_b)
        vec_c = sympy.Matrix(vec_c)

        #create transformation matrix from the input basis to a cartesian basis
        fin = fslib.read_fs_inp(opt['inp'],clean=False)
        T_i = create_Ti(fin)

        #Transform to a cartesian coordinate system
        vec_a = T_i * vec_a
        vec_b = T_i * vec_b
        vec_c = T_i * vec_c

        #calculate the dual basis vectors
        V = vec_a.dot(vec_b.cross(vec_c))
        vec_A = old_div(vec_b.cross(vec_c),V)
        vec_B = old_div(vec_c.cross(vec_a),V)
        vec_C = old_div(vec_a.cross(vec_b),V)

        if debug:

            print('basis vectors')
            print(vec_a,vec_b,vec_c)
            print('dual basis vectors')
            print(vec_A,vec_B,vec_C)

        A_l = vec_a.row_join(vec_b)    
        A_l = A_l.row_join(vec_c)

        A_L = vec_A.row_join(vec_B)    
        A_L = A_L.row_join(vec_C)

        if debug:
            print('A_l')
            print(A_l)
            print('A_L')
            print(A_L)

        if opt['num_prec'] is None:
            G = A_l.LUsolve(A_L)
        else:
            A_ln = np.array(A_l).astype(np.float64)
            A_Ln = np.array(A_L).astype(np.float64)
            if debug:
                print('A_ln')
                print(A_ln.dtype)
                print('A_Ln')
                print(A_Ln.dtype)
            G = np.linalg.solve(A_ln,A_Ln)
            G = sympy.Matrix(G)

        if debug:
            print('The metrix tensor')
            sympy.pprint(G)
            print('The inverse of the metrix tensor')
            sympy.pprint(G.inv())

        if debug:
            print('metric tensor test')
            for i in range(3):
                a_I = 0*vec_a
                for j in range(3):
                    a_I += G.T[i,j]*A_l[:,j]
                print(a_I)

            print('Get metric debug output end')
            print('================================')
            print('')

    return G


def get_generators(syms,depth=3,debug=False,prec=None):
    if not isinstance(depth,int) or depth < 1 or depth > 3:
        raise Exception('Depth must be integer between 1 and 3')
    gens = [syms[1]]
    gens_i = [1]

    def equal_function(sym1,sym2):
        if prec is None:
            return sym1 == sym2
        else:
            return sym1.eq_numeric(sym2,prec=prec)
    for i, sym in enumerate(syms):
        is_generated = False
        for gen1 in gens:
            if equal_function(sym, gen1):
                is_generated = True
                break
            if equal_function(sym, gen1.inv()):
                is_generated = True
                break
            if depth > 1:
                for gen2 in gens:
                    if equal_function(gen1 * gen2, sym):
                        is_generated = True
                        break
                    if equal_function(gen1 * gen2.inv(), sym):
                        is_generated = True
                        break
            if is_generated:
                break
        if depth > 2:
            if not is_generated:
                for gen1 in gens:
                    for gen2 in gens:
                        for gen3 in gens:
                            if equal_function(gen1 * gen2 * gen3, sym):
                                is_generated = True
        if not is_generated:
            gens.append(sym)
            gens_i.append(i)
            if debug:
                print('Adding symmetry: {}, {} to generators'.format(i,sym))
        else:
            if debug:
                print('Symmetry {} {} is generated.'.format(i,sym))

    return gens_i, gens

def remove_symmetry(syms,R_sym,mode=0):
    if not isinstance(R_sym,Symmetry):
        if R_sym == 'P':
            R_sym = create_P()
        elif R_sym == 'I':
            R_sym = create_I()
        elif R_sym == 'T':
            R_sym = create_T()
        else:
            raise Exception('Wrong R_sym format!')
    #If R_sym has no permutations we assume it doesn't permute
    #the atoms. This is correct for I and T, but not necessarily for P!!!
    if R_sym.permutations is None:
        R_sym.permutations = {}
        for key in syms[0].permutations:
            R_sym.permutations[key] = key
    syms_out_i = []
    syms_out = []
    if mode == 0:
        for i,sym in enumerate(syms):
            sym2 = sym * R_sym
            if sym2 not in syms_out:
                syms_out.append(sym)
                syms_out_i.append(i)
    elif mode == 1:
        for i,sym in enumerate(syms):
            if sym != R_sym:
                syms_out.append(R_sym)
                syms_out_i.append(i)
    return syms_out_i,syms_out

def simplify_symmetry_operations(syms,generators=True,remove_P=False,remove_T=False):

    idxs = list(range(len(syms)))
    syms_g = syms
    if remove_P:
        i1, syms_g = remove_symmetry(syms_g, 'P')
        idxs = [idxs[i] for i in i1]
    if remove_T:
        i1, syms_g = remove_symmetry(syms_g, 'T')
        idxs = [idxs[i] for i in i1]
    if generators:
        i1, syms_g = get_generators(syms_g)
        idxs = [idxs[i] for i in i1]

    return idxs,syms_g

def get_full_permutations(lines,prec=3,debug=False):
    """
    This function determines permutations for all atoms. This is necessary in cases, where
    the non-magnetic unit cell is smaller than the magnetic and then the conventional method
    only determines the permutation for a subset of atoms, which is then a problem for noso.

    The way this works is:
    1. Get a list of atoms in the input, which are given in teh findsym basis.
    2. Convert those to the crystallographic basis used in findsym output.
    3. Transform these by symmetries.
    4. Transform back to the findsym basis.
    5. Check the transformations.

    In addition, when the non-magnetic unit cell is smaller we need to add additional translations to
    the original symmetry operations. These are identified by the find_translations function and are
    added to the non-magnetic symmetry operations.

    Args:
        lines: the findsym output, as a list of lines
        prec: the rounding precision, 3 means 3 digits of rounding and then atomic positions
            closer than 1e-2 are taken as equal.

    Returns:
        out: list of lists such that out[i] contains all the permutations for the non-magnetic symmetry
        i, that is the permutation for the original non-magnetic symmetry and the permutatinos for all
        the translations
    """

    prec2 = 1e-4

    if debug:
        print('get_full_permutations starting')

    origin = np.array(fslib.r_origin(lines),dtype=float)
    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    Tm = np.array(create_Tm(vec_a,vec_b,vec_c),dtype=float)
    Tmi = np.array(np.linalg.inv(Tm),dtype=float)

    if norm(np.rint(Tmi) - Tmi) > prec2:
        raise Exception('Non-integerer transformation matrix: rotation of coordinate system present.')

    ts = find_translations(Tm)
    if debug:
        print(ts)

    start = False
    end = False
    pos_o = []
    for line in lines:
        if 'Position and magnetic moment of each atom (dimensionless coordinates)' in line:
            start = True
            continue
        if start and '------------------------------------------' in line:
            end = True
            continue
        if start and not end:
            pos_o.append([float(i) for i in line.split()[1:4]])          

    if debug:
        print('pos_o')
        for p in pos_o:
            print(p)
        print('')
    
    pos_T = []
    for p in pos_o:
        pos_T.append(np.dot(Tmi,p[0:3]-origin))

    if debug:
        print('pos_T')
        for p in pos_T:
            print(p)
        print('')
        
    syms = fslib.r_sym(lines,syms_only=False)
    
    out = []
    for isym,sym in enumerate(syms):
        if debug:
            print('')
            print(isym)
            print(sym[1])
            print(sym[4])
            print({x[0]: x[1] for x in sorted(sym[4])})
        out_t = []
        for trans in ts:
            if debug:
                print(trans)
            pos_TS = []
            for pT in pos_T:
                pTS = []
                for i in range(3):
                    pTS.append(sym[1][i])
                    for j,s in enumerate(['x','y','z']):
                        pTS[i] = pTS[i].replace(s,str(pT[j]))
                pTS = np.array(sympy.sympify(pTS),dtype=float)+trans
                pos_TS.append(pTS)

            if debug:
                print('pos_TS')
                for p in pos_TS:
                    print(p)
                print('')

            pos_TSTi = []
            for p in pos_TS:
                pos_TSTi.append( np.round( np.dot( np.array(Tm,dtype=float), p) + origin, prec) % 1)

            pos_TSTi_test = []
            for p in pos_TS:
                pos_TSTi_test.append( np.dot( np.array(Tm,dtype=float), p) + origin)

            if debug:
                print('pos_TSTi')
                for p in pos_TSTi:
                    print(p)
                print('')

            if debug:
                print('pos_TSTi_test')
                for p in pos_TSTi_test:
                    print(p)
                print('')

            pos_or = np.round(pos_o,prec) % 1
            if debug:
                print('pos_or')
                print(pos_or)

            permutations = {}
            for i,p in enumerate(pos_TSTi):
                found = False
                for j,po in enumerate(pos_or):
                    if np.linalg.norm(p-po) < 1.5/(10**(prec)):
                        #if debug:
                        #    print(i,p)
                        #    print(j,po)
                        #    print('')
                        perm = j
                        found = True
                        break
                if not found:
                    print(p)
                    print(pos_or)
                    raise Exception('Permutation not found')
                else:
                    permutations[i+1] = perm+1
            #print(sym[4])
            if debug:
                print(permutations)
            out_t.append(permutations)
        out.append(out_t)
        
    return out

def get_integer_solution(A, b, debug=False):
    """
    Finds an integer solution to a system of linear equtions.
    If there is no solution returs None.
    """
    B, U, V = smith_normal_form(A)
    if debug:
        print(B)
        print(U)
        print(V)
        print(np.dot(U, np.dot(A, V)))
    c = np.dot(U, b)
    y = np.zeros(B.shape[1])

    has_solution = True
    for i in range(B.shape[0]):
        if B[i, i] == 0:
            continue
        if c[i] % B[i, i] == 0:
            y[i] = c[i] // B[i, i]
        else:
            has_solution = False
            break
    if not has_solution:
        return None
    else:
        return np.dot(V, y)

def find_translations(T):
    """
    For matrix T that defines a supercell it finds all lattice translations that needs to be added to the symmetries
    of the supercell. This is because these translations are Bravais lattice translations in the smaller unit cell, but
    not in the supercell.

    This works by trying different combinations of the small unit cell lattice vectors and trying if these are an
    integer linear combination of the large unit cell lattice vectors (that are defined by T) or if they differ from
    already identified translation by an integer linear combination of the large unit cell lattice vectors.

    If det(T) = n that means the supercell is n-times larger and that should mean there is n translations.
    """

    prec = 1e-4
    Ti_o = np.linalg.inv(T)
    Ti = np.rint(Ti_o).astype(int)

    if norm(Ti_o - Ti) > prec:
        raise Exception('Not integer matrix!')

    det = np.rint(np.linalg.det(Ti))
    abs_det = abs(int(det))

    n_trans = 0

    it = [0, 1, -1, 2, -2, 3, -3, 4, -4, 5, -5]
    itt = list(product(it, it, it))
    itt = sorted(itt, key=norm)

    A = Ti.copy()

    ts = [np.array([0, 0, 0])]

    i = 0
    # while len(ts) < abs(det):
    while len(ts) < abs_det and i < len(itt):
        i += 1
        t = np.array(itt[i])
        # print(i)
        # print(t)

        found = False
        for ti in ts:
            x = get_integer_solution(A, t - ti)
            if x is not None:
                found = True
                break
        if not found:
            ts.append(t)
    if len(ts) < abs_det:
        print(i,len(itt))
        print(abs_det)
        print(ts)
        raise Exception('Did not find all transformations.')
    return ts

def get_full_permutations_old(lines, prec=3, debug=False):
    """
    This function determines permutations for all atoms. This is necessary in cases, where
    the non-magnetic unit cell is smaller than the magnetic and then the conventional method
    only determines the permutation for a subset of atoms, which is then a problem for noso.

    The way this works is:
    1. Get a list of atoms in the input, which are given in teh findsym basis.
    2. Convert those to the crystallographic basis used in findsym output.
    3. Transform these by symmetries.
    4. Transform back to the findsym basis.
    5. Check the transformations.

    Args:
        lines: the findsym output, as a list of lines
        prec: the rounding precision, 3 means 3 digits of rounding and then atomic positions
            closer than 1e-2 are taken as equal.

    Returns:
        list of permutations for all symmetry operations
    """

    if debug:
        print('get_full_permutations starting')

    origin = np.array(fslib.r_origin(lines), dtype=float)
    [vec_a, vec_b, vec_c] = fslib.r_basis(lines)
    Tm = create_Tm(vec_a, vec_b, vec_c)
    Tmi = np.array(Tm.inv().evalf(), dtype=float)

    start = False
    end = False
    pos_o = []
    for line in lines:
        if 'Position and magnetic moment of each atom (dimensionless coordinates)' in line:
            start = True
            continue
        if start and '------------------------------------------' in line:
            end = True
            continue
        if start and not end:
            pos_o.append([float(i) for i in line.split()[1:4]])

    if debug:
        print('pos_o')
        for p in pos_o:
            print(p)
        print('')

    pos_T = []
    for p in pos_o:
        print('wtttf')
        pos_T.append(np.dot(Tmi, p[0:3] - origin))

    if debug:
        print('pos_T')
        for p in pos_T:
            print(p)
        print('')

    syms = fslib.r_sym(lines, syms_only=False)

    out = []
    for isym, sym in enumerate(syms):
        if debug:
            print('')
            print(isym)
            print(sym[1])
            print(sym[4])
            print({x[0]: x[1] for x in sorted(sym[4])})
        pos_TS = []
        for pT in pos_T:
            pTS = []
            for i in range(3):
                pTS.append(sym[1][i])
                for j, s in enumerate(['x', 'y', 'z']):
                    pTS[i] = pTS[i].replace(s, str(pT[j]))
            pTS = np.array(sympy.sympify(pTS), dtype=float)
            pos_TS.append(pTS)

        if debug:
            print('pos_TS')
            for p in pos_TS:
                print(p)
            print('')

        pos_TSTi = []
        for p in pos_TS:
            pos_TSTi.append(np.round(np.dot(np.array(Tm, dtype=float), p), prec) % 1)

        pos_TSTi_test = []
        for p in pos_TS:
            pos_TSTi_test.append(np.dot(np.array(Tm, dtype=float), p))

        if debug:
            print('pos_TSTi')
            for p in pos_TSTi:
                print(p)
            print('')

        if debug:
            print('pos_TSTi_test')
            for p in pos_TSTi:
                print(p)
            print('')

        pos_or = np.round(pos_o, prec) % 1
        if debug:
            print('pos_or')
            print(pos_or)

        permutations = {}
        for i, p in enumerate(pos_TSTi):
            found = False
            for j, po in enumerate(pos_or):
                if np.linalg.norm(p - po) < 1 / (10 ** (prec - 1)):
                    # if debug:
                    #    print(i,p)
                    #    print(j,po)
                    #    print('')
                    perm = j
                    found = True
                    break
            if not found:
                raise Exception('Permutation not found')
            else:
                permutations[i + 1] = perm + 1
        # print(sym[4])
        if debug:
            print(permutations)
        out.append(permutations)

    return out





# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from six.moves import urllib
from lxml import html
from bs4 import BeautifulSoup
import pandas as pd
import re
from pymatgen.core import Structure, Lattice, Element

def download_html(id):
    page = urllib.request.urlopen('http://webbdcrista1.ehu.es/magndata/index.php?index='+id)
    mybytes = page.read()
    page_txt = mybytes.decode("utf-8", errors='ignore')
    page.close()
    return page_txt

def get_magndata_structure(id):
    try:
        htmlstr = download_html(id)
        root = html.fromstring(htmlstr)
        dfs = pd.read_html(process_html(htmlstr))

    except Exception as e:
        raise Exception("Error in downloading magndata page {}".format(e))

    info = {}
    # Parse html contents
    lattice_parameter = parse_lattice_parameter(root)
    lattice = Lattice.from_parameters(*lattice_parameter)

    # Fetch tabled contents
    sites, tabled_info = parse_tables(dfs)
    species = sites['species']
    coords = sites['coords']
    moments = sites['moments']
    occupancy = sites['occupancy']
    constraints = sites['constraints']
    labels = sites['labels']

    properties = {'magmom': moments, 'magmom_basis': ['crystal_scaled'] * len(moments), 'occupancy': occupancy,
                  'label': labels, 'magnetic_constraints': constraints}
    try:
        structure = Structure(lattice, species, coords, site_properties=properties)
    except Exception as e:
        raise Exception("Error in extracting the structure from MAGNDATA: {}".format(e))
    return structure


def parse_lattice_parameter(root):
    body_xpath = root.xpath('/html/body')
    body_lines = body_xpath[0].text_content().split('\n')

    for i, line in enumerate(body_lines):
        if re.search('Lattice parameters of the magnetic unit cell', line):
            lattice_line = body_lines[i + 1]
            break

    try:
        lstr = re.findall(r'([0-9]*\.[0-9]*)', lattice_line)
        lattice_parameter = [float(i) for i in lstr]
    except Exception as e:
       raise Exception("Problem in extracting lattice parameters from MAGNDATA: {}".format(e))

    return lattice_parameter

def parse_tables(dfs):
    # Extract atomic tables
    tables = []
    for df in dfs:
        column = df.columns.map(str)
        if 'Label' in column:
            tables.append(df)
        if 'Atom' in column:
            # tables.append(df[~df['Atom'].str.contains('click')])
            delete = []
            for i, row in df.iterrows():
                rowstr = row.Atom
                #if type(rowstr) is str or type(rowstr) is unicode:
                if type(rowstr) is str:
                    if not rowstr.isdigit():
                        delete.append(i)
            tables.append(df.drop(index=delete))

    # Remove duplication
    for i, table in enumerate(tables):
        column = table.columns.map(str)
        if '|M|' in column:
            start_index = i
            break
    for i in range(start_index + 1, len(tables)):
        column = tables[i].columns.map(str)
        if '|M|' in column:
            end_index = i
            break
    atom_tables = [tables[i] for i in range(start_index,end_index)]

    # Get infomations
    labels = []
    atom_types = []
    mag_atoms = []
    multiplicities = []
    coord_tables = []
    dis_atoms = []
    dis_labels = []
    dis_occupancy = {}
    dis_coord = []
    for table in atom_tables:
        column = table.columns.map(str)
        if 'Label' in column:
            labels.extend(table.loc[:,'Label'])
            atom_types.extend(table.loc[:, 'Atom type'])
            atom_types = strip_atom(atom_types)
            multiplicities.extend(table.loc[:,'Multiplicity'])
            if '|M|' in column:
                mag_atoms.extend(table.loc[:,'Label'])
            if 'Occupancy' in column:
                alloy = True
                for j, row in table.iterrows():
                    if row.Occupancy < 1:
                        dis_atoms.append(row['Atom type'])
                        dis_labels.append(row['Label'])
                        dis_occupancy[row.Label] = row.Occupancy
                        coord = [row.x,row.y,row.z]
                        dis_coord.append(coord)
            else:
                alloy = False
        elif 'Atom' in column:
            coord_tables.append(table)
        else:
            print('Error: Failed to parse MAGNDATA tables.')
            exit(1)

    # Find the positional overlapping of disorder atoms
    lapped_pairs = []
    if alloy:
        for i, label1 in enumerate(dis_labels):
            for j in range(i+1,len(dis_labels)):
                label2 = dis_labels[j]
                if dis_coord[i] == dis_coord[j]:
                    lapped_pairs.append([label1,label2])

    species = []
    ext_labels = []
    coords = []
    moments = []
    mag_constraints =[]
    occupancy = []
    for i, table in enumerate(coord_tables):
        column = table.columns.map(str)

        for j, row in table.iterrows():
            species.append(atom_types[i])
            ext_labels.append(labels[i])
            coord = [row.x,row.y,row.z]
            coords.append(convert_float(coord))
            if labels[i] in mag_atoms:
                moment = [row.Mx,row.My,row.Mz]
                moments.append(convert_float(moment))
                mag_constraints.append(row['Symmetry constraints on M'])
            else:
                moments.append([0.0] * 3)
                mag_constraints.append('0,0,0')
            if alloy:
                if labels[i] in dis_labels:
                    try:
                        occupancy.append(float(dis_occupancy[labels[i]]))
                    except:
                        print_exc()
                        exit(3)
                else:
                    occupancy.append(1.0)
            else:
                occupancy.append(1.0)

    elements = list(set(atom_types))
    max_elem_num = max_element_number(elements)

    num = float(sum(occupancy))
    if num.is_integer():
        defect = False
    else:
        defect = True

    sites = {
        'labels': ext_labels,
        'species': species,
        'coords': coords,
        'moments': moments,
        'occupancy': occupancy,
        'constraints': mag_constraints
    }

    info = {
        'num_of_atoms': num,
        'overlapped_pairs': lapped_pairs,
        'alloy': alloy,
        'defect': defect,
        'magnetic_atoms': mag_atoms,
        'disorder_atoms': dis_atoms,
        'elements': elements,
        'max_element_number': max_elem_num,
        }

    return sites, info

def process_html(htmlstr):
    # Merge the sepalate tables in MAGNDATA html (for pandas.read_html)

    # Decompose merged cells
    new = re.sub(r'<tr><td[^>]*colspan=[^>]*>.*?<\/td><\/tr>','',htmlstr)

    # Fix HTML
    soup = BeautifulSoup(new, "lxml")
    for body in soup("tbody"):
        body.unwrap()

    return str(soup)

def convert_float(vec):
    try:
        newvec = [float(x) for x in vec]
        return newvec
    except Exception as e:
        raise Exception("Problem in convert_float {}".format(e))

def strip_atom(atoms):
    spec = []
    for atom in atoms:
        a = re.search(r'[a-z|A-Z]+', atom).group()
        # Deuteriums are converted into simple hydrogen
        if a == 'D':
            a = 'H'
        spec.append(a)
    return spec

def max_element_number(elements):
    elements_nums = [Element(e).number for e in elements]
    return max(elements_nums)



from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import range
from builtins import object
import re
import copy
import sys
import math
import time

import sympy
from scipy.linalg import lu,qr
from scipy.io import savemat
import numpy as np
import numpy.linalg
from numpy.linalg import svd
import mpmath   

from .tensors import matrix, mat2ten, tensor2Y
from .fslib import transform_position
from .conv_index import *

def num_rref(Y,prec=15):

    Y = np.matrix(Y)

    pl,U = lu(Y,permute_l=True)
    #P,L,U = mpmath.lu(Y)

    #print('L,U decomposition')
    #print(pl)
    #print(U)

    pivots = []
    for i in range(U.shape[0]):
        for j in range(U.shape[1]):
            pivot_found = False
            if U[i,j] != 0 and not pivot_found:
                pivots.append(j)
                U[i,:] /= U[i,j]
                break
    try:
        np.linalg.inv(pl)
    except:
        print('The PL matrix is singular. The output of the code may be wrong!!!!')
    return U,pivots

def QR_rref(V,num_prec=1e-10):
    Q,R = qr(V)
    return U_to_rref(R,num_prec=num_prec)

def U_to_rref(U,num_prec=1e-10):
    U_out = U.copy()
    pivots = []
    for row in range(U_out.shape[0]):
        found_pivot = False
        for col in range(U_out.shape[1]):
            if abs(U_out[row,col]) >= num_prec:
                if col not in pivots:
                    pivots.append(col)
                    U_out[row,:] = U_out[row,:] / U_out[row,col]
                    found_pivot = True
                    break
                else:
                    U_out[row,:] -= U_out[pivots.index(col),:] * U_out[row,col]
        if not found_pivot:
            raise Exception('Did not find pivot!')
            
    pivots_i = [(p,i) for i,p in enumerate(pivots)]
    pivots_i.sort()
    sorted_pivots,permutation = zip(*pivots_i)
    if permutation != tuple(range(len(pivots))):
        U_out_perm = U_out.copy()
        for row in range(U_out.shape[0]):
            U_out_perm[row,:] = U_out[permutation[row],:]
    
        pivots = sorted_pivots
        U_out = U_out_perm

    for piv_row,col in enumerate(pivots):
        for row in range(U_out.shape[0]):
            if row < piv_row:
                if abs(U_out[row,col]) >= num_prec:
                    U_out[row,:] -= U_out[row,col] * U_out[piv_row,:]
    return U_out,pivots

def get_rref(U,num_prec=1e-10):
    """
    This transforms the matrix in to a reduced row echelong form. The algorithm is taken
    from octave and should be exactly the same.

    Columns that are smaller than num_prec are set to zero.
    """
    U_o = U.copy()
    pivots = []
    row = 0
    rows,columns = U.shape
    for col in range(columns):
        m,pivot = [np.max(np.abs(U_o[row:,col])),np.argmax(np.abs(U_o[row:,col]))]
        pivot = pivot + row
        
        if m < num_prec:
            U_o[row:,col] = np.zeros((rows-row))
        else:
            pivots.append(col)
            U_o[[pivot,row],:] = U_o[[row,pivot],:]
            U_o[row,:] = U_o[row,:] / U_o[row,col]
            for r in range(rows):
                if r != row:
                    U_o[r,:] -= U_o[row,:] * U_o[r,col]
            if row == rows-1:
                break
            else:
                row += 1
    
    return U_o,pivots

class params_trans(object):
    def __init__(self,op1,op2,op3,l,T=None,sym_format='findsym'):
        self.op1 = op1
        self.op2 = op2
        self.op3 = op3
        self.l = l
        self.T = T
        self.sym_format = sym_format

class SymmetrOpt(object):
    def __init__(self,num_prec=None,debug=False,debug_time=False,debug_Y=False,round_prec=None,numX=False):
        self.num_prec = num_prec
        self.debug = debug
        self.debug_time = debug_time
        self.debug_Y = debug_Y
        self.round_prec = round_prec
        self.numX = numX

def symmetr(syms,X,trans_func,params,opt=None):
    """
    This symmetrizes a tensor X given a list of symmetries and a transformation function.

    This function should be quite general and is now used for all symmetrizing.

    Args:
        syms: list of symmetry operations
        X: tensor - must be a tensor class
        trans_func: function that transforms the tensor X using symmetry sym
            trans_func must work in the following way:
            X_trans = trans_func(X,sym,params)
            If trans_func returns None then the symmetry operation is ignored
        params: parameters to be sent to function trans_func

    Returns:
        X_trans: the symmetry restricted form of tensor X
    """
    if opt is None:
        opt = SymmetrOpt()
    num_prec = opt.num_prec
    debug = opt.debug
    debug_time = opt.debug_time
    debug_Y = opt.debug_Y
    if opt.numX and num_prec is None:
        raise Exception('In the numX mode, num_prec must be specified!')

    if debug_time:
        print('Symmetrize starting')

    if num_prec is None:
        algo = 1
        algo_solve = 0
    else:
        algo = 3
        algo_solve = 1

    if debug:
        print('')
        print('======= Starting symmetrizing =======')

    #we do a loop over all symmetry operations, for each symmetry, we find what form the response matrix can have, when the system has this symmetry
    #for next symmetry we take the symmetrized matrix from the previous symmetry as a starting point
    for sym in syms:
        
        if debug:
            print('Symmetry:') 
            print(sym)
            print('')

        if debug_time:
            t0 = time.perf_counter()

        X_trans = trans_func(X,sym,params)
        if X_trans is None:
            continue

        if debug:
            print('')
            print('Current form of the tensor:')
            print('')
            X.pprint()
            print('')
            print('Transformed tensor:')
            print('')
            X_trans.pprint()
        #X.pprint()
        #X_trans.pprint()
        #The tensor must be equal to the transformed tensor, this give us a system of linear equations.
        #matrix Y is a matrix that represents this system, ie the system X-X_trans = 0
        #we reverse the order of the rows
        # it doesn't really matter but the results are more natural this way

        if debug_time:
            t1 = time.perf_counter()
            print('Time for transforming the tensor: ',t1-t0) 
            t0 = time.perf_counter()

        #equal = (X_trans == X)
        if debug_time:
            t3= time.perf_counter()
            print('Time for checking equality: ',t3-t0)
        #if equal:
        #    continue

        if not opt.numX:
            Y = tensor2Y(X-X_trans,Y_format='sympy',algo=algo,reverse=True,td=False)
        else:
            Y = X.t - X_trans.t

        """
        algo_solve == 0: We use the sympy rref function. This seems to have issues though even if num_prec is set.
        algo_solve == 1: We svd decomposition to find the nullspace and the convert to rref using get_rref.
        """
        if algo_solve == 0:

            #this transforms the matrix into the Reduced row echelon form
            #piv are the indeces o the pivot columns
            if num_prec is None:
                [rref,piv] = Y.rref()
            else:
                def zerofunc(x):
                    try:
                        a = abs(x) < num_prec
                        if a:
                            result = True
                        else:
                            result = False
                    except:
                        result = None
                    #print(result)
                    return result
                #sympy.pprint(Y)
                [rref,piv] = Y.rref(iszerofunc=lambda x:abs(x)<num_prec)        
                #[rref,piv ] = num_rref(Y)
                #[rref,piv] = Y.rref(iszerofunc=zerofunc)        

            if debug_time:
                t2 = time.perf_counter()
                print('Time for reducing Y to reduced row echelon form: ', t2-t1)

            if debug_Y:
                print('')
                print('Matrix representing the linear equation system that has to be satisfied: (right hand side is zero)')
                sympy.pprint(Y)
                print('')
                print('Reduced row echelon form and indeces of the pivot columns:')
                sympy.pprint(rref)
                print(piv)
                print('')

            #a loop over all the pivots: it's the pivots that give interesting information
            for j in list(reversed(piv)):

                
                #find the row of pivot j
                found = False
                i = X.dim1**X.dim2-1
                while found == False:
                    if rref[i,j] == 1:
                        found = True
                    else:
                        i = i-1
                
                if debug:
                    print('')
                    print('considering pivot ', i,j)

                tmp = 0
                #now we just make use of the linear equation that holds for this pivot
                #keep in mind that the rows are in reversed order
                for ll in range(j+1,X.dim1**X.dim2):
                    tmp = tmp - rref[i,ll]*X.x[rev_inds[ll]]
                X = X.subs(X.x[rev_inds[j]],tmp)

                if debug:
                    print('substituting ', end=' ')
                    sympy.pprint(X.x[rev_inds[j]])
                    print(' for ', end=' ')
                    sympy.pprint(tmp)
                    print('')

        elif algo_solve == 1:
            """
            We use the svd decomposition to find the nullspace, this is given by V2:
            rows of V2 are vectors which form the basis of the Y nullspace. This means that 
            any combination of these vectors is a solution of Yx = 0.

            Therefore any solution of Yx =0, must be written as:
            x = a_i * v_i,
            where v_i are the rows of V2. 
            By transforming V2 into the reduced row echelon form, we can then
            eliminate some variables from X. The information is extracted from columns of the V2_rref matrix
            (nor rows like in the case of Y_rref). The pivot rows simply tell us that a_i = x_i, the non-pivot
            columns then give the relation of the other x_i's in terms of the pivot x_i's.
            """

            if debug_time:
                ts0 = time.perf_counter()

            if opt.numX:
                U,S,V = svd(Y)
            else:
                U, S, V = svd(np.array(np.flip(Y, axis=1), dtype=float))
            if debug_time:
                ts1 = time.perf_counter()
                print('Time for svd: {}'.format(ts1-ts0))
            zero_singulars = [i for i,s in enumerate(S) if abs(s) < num_prec]

            #if there are no singular values, then the tesor has to be zero:
            #thus we don't have to do any substituting
            if len(zero_singulars) == 0:
                X = X.copy0()
            else:
                #V2 = sympy.zeros(len(zero_singulars),Y.shape[1])
                V2 = np.zeros((len(zero_singulars),Y.shape[1]))
                for i in range(len(zero_singulars)):
                    V2[i,:] = V[zero_singulars[i],:]
                if debug_Y:
                    print('Singular values: ',S)
                    print(V2)
                V2_rref,pivs = get_rref(V2,num_prec)
                if debug_Y:
                    print(V2_rref)
                if debug_time:
                    ts0 = time.perf_counter()
                    print('Time for rref: {}'.format(ts0-ts1))

                #This is a consistency check: the rows of V2 should be linearly indpenedent and so
                #there should be no zero rows in V2_rref
                if len(pivs) != len(zero_singulars):
                    Vn = np.array(V2.evalf(),dtype=float)
                    np.save('debug_V.npy',Vn)
                    print(pivs)
                    print(zero_singulars)
                    print([S[i] for i in zero_singulars])
                    for i in range(len(zero_singulars)):
                        for j in range(Y.shape[1]):
                            V2[i,j] = round(V2[i,j],3)
                            V2_rref[i,j] = round(V2_rref[i,j],3)

                    sympy.pprint(V2)
                    sympy.pprint(V2_rref)
                    raise Warning('Issue with rref of V2, results are likely to be wrong!!!')

                if not opt.numX:
                    ais = []
                    for i in range(len(zero_singulars)):
                        ais.append(X.x[X.inds[pivs[i]]])

                    for j in range(Y.shape[1]):
                        if j not in pivs:
                            tmp = 0
                            print_debug = False
                            for i in range(len(zero_singulars)):
                                coeff = round(V2_rref[i,j],opt.round_prec)
                                tmp += coeff * ais[i]
                                #tmp += V2_rref[i,j] * ais[i]
                                if abs(coeff) > 10 or (abs(coeff) < 0.1 and abs(coeff) > 1e-3):
                                    print('Warning: Large or small coefficient {}, this can signify error'.format(coeff))
                                    print_debug = True
                            if debug or print_debug:
                                print('Substituting {} -> {}'.format(X.x[X.inds[j]],tmp))
                            X = X.subs(X.x[X.inds[j]],tmp)
                else:
                    for j in range(Y.shape[1]):
                        if j not in pivs:
                            sub_vec = np.zeros(Y.shape[1])
                            for i in range(len(zero_singulars)):
                                sub_vec[pivs[i]] = V2_rref[i,j]
                            X.subs(j,sub_vec)

            if debug_time:
                ts1 = time.perf_counter()
                print('Time for subs: {}'.format(ts1-ts0))
        else:
            raise Exception('Unknown algo_solve')

        #if debug_time:
        #    t1 = time.perf_counter()
        #    print('Time for constructing tensor ', t1-t2)
        #    print('')


        if debug:
            print('Current form of the tensor:')
            X.pprint()
            print('')

    if opt.round_prec is not None:
        X.round(opt.round_prec)
        
    if debug:
        print('Symmetrized tensor:')
        X.pprint()
        print('')
        print('======= End symmetrizing =======')

    return X

def even_odd(Xs):
    """Finds whether the first part of the response tensor is even or odd
    
    Args:
        op1,op2,op3: the operator types
        Returns: either ('even','odd') or ('odd','even')
    """
    if Xs[0].is_even() and not Xs[1].is_even():
        return('even','odd')
    elif not Xs[0].is_even() and Xs[1].is_even():
        return('odd','even')
    else:
        print(Xs[0].is_even(),Xs[1].is_even())
        raise Exception('Wrong transformation under time-reversal')


def symmetrize_res(symmetries,X,proj=-1,s_opt=None):

    syms_sel = []
    for sym in symmetries:
        
        if s_opt.debug:
            print('Symmetry:') 
            print(sym)
            print('')
            if proj != -1:
                print('Symmetry transforms the atom ', proj, ' into atom ', sym.permutations[proj])
                if  sym.permutations[proj] != proj:
                    print('Skipping symmetry')
                    print('')

        #if there is a projection set up we only consider symmetries that keep the atom invariant
        if proj == -1 :
            take_sym = True
        elif sym.permutations[proj] == proj:
            take_sym = True
        else:
            take_sym = False

        if take_sym:
            syms_sel.append(sym)

    def trans_func(X,sym,params):
        return X.transform(sym)
    X = symmetr(syms_sel,X,trans_func,None,s_opt)

    return X

def symmetrize_same_op(X,s_opt=None):

    perm = {}
    for i in range(X.dim2):
        perm[i] = i
    perm[0] = 1
    perm[1] = 0
    perms=[perm] 

    def trans_func(X,perm,params):
        X_T = X.copy0()

        for ind in X:
            ind_T = [0]*len(ind)
            for i in range(len(ind)):
                ind_T[i] = ind[perm[i]]
            ind_T = tuple(ind_T)
            X_T[ind_T] = X.T_comp * X[ind]

        ind_types = [0] * X.dim2
        for i in range(X.dim2):
            ind_types[i] = X.ind_types[perm[i]]
        X_T.ind_types = tuple(ind_types)

        for i in range(X.dim2):
            if X_T.ind_types[i] != X.ind_types[i]:
                X_T.reverse_index(i)

        return X_T

    X = symmetr(perms,X,trans_func,None,s_opt)
        
    return X

def symmetrize_sym_inds(X,sym_inds,asym_inds,s_opt=None):

    def trans_func(X,perm,sign):

        X_T = X.copy0()

        for ind in X:
            ind_T = [0]*len(ind)
            for i in range(len(ind)):
                ind_T[i] = ind[perm[i]]
            ind_T = tuple(ind_T)
            X_T[ind_T] = sign * X[ind]

        ind_types = [0] * X.dim2
        for i in range(X.dim2):
            ind_types[i] = X.ind_types[perm[i]]
        X_T.ind_types = tuple(ind_types)

        for i in range(X.dim2):
            if X_T.ind_types[i] != X.ind_types[i]:
                X_T.reverse_index(i)

        return X_T

    if sym_inds is not None:

        perms_sym = []
        for si in sym_inds:
            perm = list(range(X.dim2))
            perm[si[0]] = si[1]
            perm[si[1]] = si[0]
            perms_sym.append(perm)
        X = symmetr(perms_sym,X,trans_func,1,s_opt)

    if asym_inds is not None:

        perms_asym = []
        for si in asym_inds:
            perm = list(range(X.dim2))
            perm[si[0]] = si[1]
            perm[si[1]] = si[0]
            perms_asym.append(perm)

        X = symmetr(perms_asym,X,trans_func,-1,s_opt)

    return X

def symmetr_AB(syms,X,atom1,atom2,round_prec=None):
    """
    Tries to transform the tensor projected on one atom to a different atom

    Args:
        syms: The symmmetry operations. Format as outputted by read.py
        X: The input tensor.
        op1: The first operator.
        op2: The second operator.
        atom1: The atom on which X is projected.
        atom2: The atom on which X is transformed.
        T (Optional[matrix]): Coordinate transformation matrix. If it is set, the symmetry operations will be transformed by this matrix.
            Symmetry operations are given in basis A. T transforms from A to B, ie Tx_A = x_B.

    Returns:
        X_trans: The transformed tensor.
    """

    X_trans = []

    found = False
    for sym in syms:
        #there will usually be more symmetries that transform from atom1 to atom2, we need only one, as they all
        #give the same results
        if sym.permutations[atom1] == atom2 and not found:
            found = True
            for l in range(2):
                X_trans.append(X[l].transform(sym))

    if found:
        if round_prec is not None:
            for X in X_trans:
                X.round(round_prec)
        return X_trans
    else:
        return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Module for finding a symmetrical form of classical magnetic Hamiltonians.
"""
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import object
import sympy as sp

from .symmetrize import symmetr,SymmetrOpt
from .tensors import Tensor,NumTensor

#testing
import sys
import os
import subprocess

import sympy

def convert_mag_ham(H,T):
    """Converts the tensor representing the magnetic Hamiltonian to a different basis.

    Note: This is just a tensor transformation, there are several other functions very
    similar to this. The only difference is which indeces transform by T and which by
    T^-1. This just corresponds to covariant and contravariant indeces and this would be
    nice to eventually implement in the code.

    Args:
        H (Tensor): the magnetic Hamiltonian in basis A
        T (matrix): . If it is set, the symmetry operations
            will be transformed by this matrix. T transforms from A to B, ie Tx_A = x_B.
    Returns:
        H_T: the magnetic Hamiltonian in basis B
    """

    Ti = T.inv()
    H_T = H.copy0()
    for ind1 in H:
        for ind2 in H_T:
            factor = 1
            for i in range(H.dim2):
                    factor *= Ti[ind1[i],ind2[i]]
            H_T[ind1] += factor*H[ind2]
    return H_T

class params_trans_ham(object):
    """Class for storing parameters of the transformation function."""
    def __init__(self,sites,debug=None,T=None,check_sym=False):
        self.sites = sites
        self.debug = debug
        self.T = T
        self.check_sym = check_sym

def trans_sites(sites,sym):
    sites_t = []
    perms = sym.permutations
    for s in sites:
        sites_t.append(perms[s])
    return sites_t

def trans_mag_ham(H,sym,params):
    """Transforms the magnetic Hamiltonian by a symmetry operation.

    There are two different modes depending on params.check_sym
    If params.chech_sym == True:
        Returns None for any symmetry operation, which transforms sites into a differet
        set of sites. If sym transform sites into the same set of sites then it returns
        transformed H for the original sites. That is:
        if sites = 1,2 and sym(sites) = 2,1 then sym is used and returned H is for
        sites = 1,2.
    If params.check_sym == False:
        Returns transformed H for the transformed sites. That is if sites = 1,2 
        and sym(sites) = 2,1 the H is returned for 2,1 sites. If sym(sites) = 2,3 then
        H is returned for 2,3 sites.

    Args:
        H (class tensor): the magnetic Hamiltonian.
        sym (): symmetry operation in the findsym format  
        params (class params_trans_ham): contains parameters for the transformation
    Returns:
        H_T (class tensor): transformed Hamiltonian
    """

    # this checks whether the symmetry operation leaves the atomic sites invariant
    # if it does then we use it, otherwise we skip it and so we return the unchanged
    # Hamiltonian
    sites = params.sites
    sites_t = trans_sites(sites,sym)
    if params.check_sym:
        if sorted(sites) != sorted(sites_t):
            return None

    R = sym.get_R('s') # select the matrix corresponding the spin transformation
    RiT = R.inv().T
    if params.debug:
        print('symmetry:')
        print(sym)
        print('symmetry in the matrix form')
        sp.pprint(R)

    if params.check_sym:
        # find how does the symmetry operation transforms the atomic sites
        sites_un = set(sites)
        perm = [0]*len(sites)
        pos_s = {}
        pos_st = {}
        for s in sites_un:
            pos_s[s] = [j for j,x in enumerate(sites) if x==s]
            pos_st[s] = [j for j,x in enumerate(sites_t) if x==s]
        for s in sites_un:
            for i,p in enumerate(pos_s[s]):
                perm[p] = pos_st[s][i] 

    if params.debug:
        print('sites:')
        print(sites)
        print('transformed sites:')
        print(sites_t)
        print('the permutation:')
        print(perm)

    # permutes the tensor index according to the permutatin of the atoms
    def perm_ind(ind,perm):
        ind_p = [0]*len(ind)
        for i,ii in enumerate(ind):
            ind_p[perm[i]] = ii
        return tuple(ind_p)

    # transforms the tensor, taking the permutatin of atoms into account
    H_t = H.copy0()
    for ind in H_t:
        for ind2 in H_t:
            factor = 1
            for i in range(len(ind)):
                factor *= RiT[ind[i],ind2[i]]
            if params.check_sym:
                ind_n = perm_ind(ind,perm)
            else:
                ind_n = ind
            H_t[ind_n] += factor*H[ind2]

    if params.debug:
        print('original tensor:')
        print(H)
        print('transformed tensor:')
        print(H_t)

    return H_t

def find_perms(sites):
    """Find permutations of identical atomic sites.

    The Hamiltonian must be invariant under interchanging atomic sites if
    they are the same. This function takes the list of sites and returns
    such permutations.
    Args:
        sites (list): list of atomic sites
    Returns:
        perms : list of permutations
    """
    perms = []
    sites_un = set(sites)
    pos_s = {}
    for s in sites_un:
        pos_s[s] = [j for j,x in enumerate(sites) if x==s]
        if len(pos_s[s]) > 1:
            for i,x in enumerate(pos_s[s]):
                for j in range(1,len(pos_s[s])):
                    if x != pos_s[s][j]:
                        perms.append((x,pos_s[s][j]))
    return perms

def trans_mag_Ham_perms(H,perm,params):
    """Transforms the Hamiltonian by a permutation of atomic sites.
    
    Args:
        H (class tensor): the magnetic Hamiltonian.
        perm (tuple): the permutation
        params: not used right now
    Returns:
        H_T: the transformed Hamiltonian
    
    """
    H_t = H.copy0()
    for ind in H:
        ind_t = list(ind)
        ind_t[perm[0]] = ind[perm[1]]
        ind_t[perm[1]] = ind[perm[0]]
        H_t[ind] = H[tuple(ind_t)]

    return H_t

def equiv(H,sites,syms,T=None,debug=False):
    """Transform H for one set of sites to all sites that are related by symmetry operation

    Args:
        H (class tensor): the magnetic Hamiltonian.
        sites (list): list of atomic sites
        syms (): symmetry operations in the findsym format
        T (sympy matrix): Coordinate transformation matrix used to transform the symmetry operations.
            Must transform syms to the coordinate system in which H is given.
    Returns:
        Asites (dictionary): the keys are all the equivalent sets of sites. Asites[sites] is the
            transformed Hamiltonian for the sites.
    """

    Asites = {}
    Asites[tuple(sorted(sites))] = H

    for sym in syms:
        sites_t = trans_sites(sites,sym)
        sites_th = tuple(sorted(sites_t))
        if sites_th not in Asites:
            params = params_trans_ham(sites,False,None)
            H_t = trans_mag_ham(H,sym,params)
            Asites[sites_th] = H_t

    return Asites

def sym_mag_ham(sites,syms,T=None,s_opt=None):
    """Returns the symmetrized magnetic Hamiltonian

    Args:
        sites (list): The atomic sites for which the Hamiltonian is considered
        syms : the symmetry operations.
        T (optinal[matrix]): Transformation matrix to a different basis. If set,
            it is used to transform the symmetry operations.
    Returns:
        The symmetrized form of the Hamiltonian
    """

    order = len(sites)
    ind_types = (-1,)*len(sites)
    H = Tensor('s', 3, order, ind_types=ind_types)

    params = params_trans_ham(sites,debug=s_opt.debug,T=T,check_sym=True)
    # first symmetrize with normal symmetry operation
    Hs = symmetr(syms,H,trans_mag_ham,params,s_opt)
    # The Hamiltonian must also be symmetric under any permutation of identical atoms
    perms = find_perms(sites)
    Hs = symmetr(perms,Hs,trans_mag_Ham_perms,params,s_opt)

    return Hs

def print_Ham(H,sites,latex=False,ret=False):
    """Prints the magnetic Hamiltonian in a nice format.
    """
    sites_un = set(sites)
    sp_vars = {}
    Hsymb = 0
    for site in sites_un:
        for i in range(3):
            comps = ['x','y','z']
            varname = 'M_'+str(site)+comps[i]
            sp_vars[(site,i)] = sp.symbols(varname)
    for ind in H:
        Hp = H[ind]
        for i,x in enumerate(ind):
            Hp *= sp_vars[(sites[i],x)]
        Hsymb += Hp

    if not latex:
        sp.pprint(Hsymb)
    else:
        print(sp.latex(Hsymb))

    if ret:
        return Hsymb



import json
import numpy as np
from symmetr import parse, sym_mham
from symmetr.tensors import Tensor

def test_mham23():
    opt = parse('mham -s 2,3 -f IrMn3/findsym.in' )
    X = sym_mham(opt)

    with open('IrMn3/mham23.out') as f:
        res = json.load(f)
    Xr = Tensor.load(res['X'])
    X_zero = Tensor(0,3,2)

    assert X - Xr == X_zero



from hsnf.Z_module import smith_normal_form

import numpy as np


def solve_frobenius_congruent(A, b=None, denominator=1000000):
    """
    solve Ax=b (mod Z^n)

    Parameters
    ----------
    A: array, (m, n)
    b: (Optional) array, (m, )
    denominator: (Optional), int

    Returns
    -------
    basis_Z: array, (rank, n)
    basis_R: array, (n - rank, n)
    x_special: array, (n, )

    general solution is written by
        x = x_special + (Z times basis_Z[0] + ...) + (R times basis_R[0] + ...)
    """
    D, P, Q = smith_normal_form(A)
    rank = np.count_nonzero(np.diagonal(D))

    D_pinv = np.zeros((Q.shape[1], rank))
    D_pinv[np.diag_indices(rank)] = 1 / D.diagonal()[:rank]
    # basis_Z[i, :] is the i-th general Z-solution of x
    basis_Z = np.dot(Q, D_pinv).T

    if rank < Q.shape[1]:
        y_R = np.zeros((Q.shape[1], Q.shape[1] - rank))
        for i in range(Q.shape[1] - rank):
            y_R[rank + i, i] = 1
        basis_R = np.dot(Q, y_R).T
    else:
        basis_R = None

    if b is None:
        return basis_Z, basis_R
    else:
        v = remainder1_with_denominator(np.dot(P, b), denominator)  # for avoiding numerical error
        if np.count_nonzero(v[rank:]) != 0:
            return basis_Z, basis_R, None

        y_special = np.zeros(A.shape[1])
        y_special[:rank] = v[:rank] / D.diagonal()[:rank]
        x_special = np.dot(Q, y_special)
        return basis_Z, basis_R, x_special


def remainder1_with_denominator(arr, denominator):
    """
    return arr (mod 1)
    """
    arr_int = np.around(arr * denominator).astype(int)
    arr_int_mod = np.remainder(arr_int, denominator)
    arr_mod = arr_int_mod / denominator
    return arr_mod



# Copyright (c) 2019 Kohei Shinohara
# Distributed under the terms of the MIT License.
import numpy as np
from hsnf.utils import (
    get_nonzero_min_abs_full,
    get_nonzero_min_abs_row,
)


class ZmoduleHomomorphism:
    """
    homomorphism between Z-modules

    Parameters
    ----------
    A: array, (m, n)
        matrix representation of homomorhism: Z^m -> Z^n
    basis_from: array, (m, )
        basis of Z^m
    basis_to: array, (n, )
        basis of Z^n
    """
    def __init__(self, A, basis_from, basis_to):
        self._A = A
        self._basis_from = basis_from
        self._basis_to = basis_to

    @property
    def num_row(self):
        return self._A.shape[0]

    @property
    def num_column(self):
        return self._A.shape[1]

    def _swap_from(self, axis1, axis2):
        self._basis_from[[axis1, axis2]] = self._basis_from[[axis2, axis1]]
        self._A[[axis1, axis2]] = self._A[[axis2, axis1]]

    def _swap_to(self, axis1, axis2):
        self._basis_to[:, [axis1, axis2]] = self._basis_to[:, [axis2, axis1]]
        self._A[:, [axis1, axis2]] = self._A[:, [axis2, axis1]]

    def _change_sign_from(self, axis):
        self._basis_from[axis] *= -1
        self._A[axis, :] *= -1

    def _change_sign_to(self, axis):
        self._basis_to[:, axis] *= -1
        self._A[:, axis] *= -1

    def _add_from(self, axis1, axis2, k):
        """
        add k times axis2 to axis1
        """
        self._basis_from[axis1] += self._basis_from[axis2] * k
        self._A[axis1, :] += self._A[axis2, :] * k

    def _add_to(self, axis1, axis2, k):
        """
        add k times axis2 to axis1
        """
        self._basis_to[:, axis1] += self._basis_to[:, axis2] * k
        self._A[:, axis1] += self._A[:, axis2] * k

    def _is_lone(self, s):
        """
        check if all s-th row elements column elements become zero
        """
        if np.nonzero(self._A[s, (s + 1):])[0].size != 0:
            return False
        if np.nonzero(self._A[(s + 1):, s])[0].size != 0:
            return False
        return True

    def _get_nextentry(self, s):
        """
        return entry which is not diviable by A[s, s]
        assume A[s, s] is not zero.
        """
        for i in range(s + 1, self.num_row):
            for j in range(s + 1, self.num_column):
                if self._A[i, j] % self._A[s, s] != 0:
                    return i, j
        return None

    def _snf(self, s):
        """
        determine SNF up to the s-th row and column elements
        """
        if s == min(self._A.shape):
            return self._A, self._basis_from, self._basis_to

        # choose a pivot
        row, col = get_nonzero_min_abs_full(self._A, s)
        if col is None:
            # if there does not remain non-zero elements, this procesure ends.
            return self._A, self._basis_from, self._basis_to
        self._swap_from(s, row)
        self._swap_to(s, col)

        # eliminate the s-th column entries
        for i in range(s + 1, self.num_row):
            if self._A[i, s] != 0:
                k = self._A[i, s] // self._A[s, s]
                self._add_from(i, s, -k)

        # eliminate the s-th row entries
        for j in range(s + 1, self.num_column):
            if self._A[s, j] != 0:
                k = self._A[s, j] // self._A[s, s]
                self._add_to(j, s, -k)

        # if there does not remain non-zero element in s-th row and column, find a next entry
        if self._is_lone(s):
            res = self._get_nextentry(s)
            if res:
                i, j = res
                self._add_from(s, i, 1)
                return self._snf(s)
            elif self._A[s, s] < 0:
                self._change_sign_from(s)
            return self._snf(s + 1)
        else:
            return self._snf(s)

    def smith_normal_form(self):
        """
        calculate Smith normal form

        see the following awesome post for a description of this algorithm:
            http://www.dlfer.xyz/post/2016-10-27-smith-normal-form/

        Returns
        -------
        D: array, (m, n)
        L: array, (m, m)
        R: array, (n, n)
            D = np.dot(L, np.dot(M, R))
            L, R are unimodular.
        """
        A = self._A.copy()
        basis_from = self._basis_from.copy()
        basis_to = self._basis_to.copy()

        D, L, R = self._snf(s=0)

        # revert A, basis_from, and basis_to
        self._A = A
        self._basis_from = basis_from
        self._basis_to = basis_to

        return D, L, R

    def _hnf_row(self, si, sj):
        """
        determine row-style HNF up to the si-th row and the sj-th column elements
        """
        if (si == self.num_row) or (sj == self.num_column):
            return self._A, self._basis_from

        # choose a pivot
        row, _ = get_nonzero_min_abs_row(self._A, si, sj)

        if row is None:
            # if there does not remain non-zero elements, go to a next column
            return self._hnf_row(si, sj + 1)
        self._swap_from(si, row)

        # eliminate the s-th column entries
        for i in range(si + 1, self.num_row):
            if self._A[i, sj] != 0:
                k = self._A[i, sj] // self._A[si, sj]
                self._add_from(i, si, -k)

        # if there does not remain non-zero element in s-th row, find a next entry
        if np.count_nonzero(self._A[(si + 1):, sj]) == 0:
            if self._A[si, sj] < 0:
                self._change_sign_from(si)

            if self._A[si, sj] != 0:
                for i in range(si):
                    k = self._A[i, sj] // self._A[si, sj]
                    if k != 0:
                        self._add_from(i, si, -k)

            return self._hnf_row(si + 1, sj + 1)
        else:
            return self._hnf_row(si, sj)

    def hermite_normal_form(self):
        """
        calculate row-style Hermite normal form

        Returns
        -------
        H: array, (m, n)
            Hermite normal form of M, upper-triangular integer matrix
        L: array, (m, m)
            unimodular matrix s.t. H = np.dot(L, M)
        """
        A = self._A.copy()
        basis_from = self._basis_from.copy()
        basis_to = self._basis_to.copy()

        H, L = self._hnf_row(si=0, sj=0)

        # revert A, basis_from, and basis_to
        self._A = A
        self._basis_from = basis_from
        self._basis_to = basis_to

        return H, L

    @classmethod
    def _standard_basis(cls, n):
        return np.eye(n, dtype=int)

    @classmethod
    def with_standard_basis(cls, A):
        """
        create homomorhism with regard A as a matrix representation with standard basis

        Parameters
        ----------
        A: array, (m, n)
            matrix representation of homomorhism: Z^m -> Z^n
        """
        A = np.array(A, dtype=int)
        if A.ndim != 2:
            raise ValueError("matrix representation must be 2d")

        m, n = A.shape
        basis_from = cls._standard_basis(m)
        basis_to = cls._standard_basis(n)

        return cls(A, basis_from, basis_to)


def smith_normal_form(M):
    """
    calculate Smith normal form

    Parameters
    ----------
    M: array, (m, n)
        integer matrix

    Returns
    -------
    D: array, (m, n)
    L: array, (m, m)
    R: array, (n, n)
        D = np.dot(L, np.dot(M, R))
        L, R are unimodular.
    """
    zmh = ZmoduleHomomorphism.with_standard_basis(M)
    return zmh.smith_normal_form()


def row_style_hermite_normal_form(M):
    """
    calculate row-style Hermite normal form
    H = LM

    Parameters
    ----------
    M: array, (m, n)

    Returns
    -------
    H: array, (m, n)
        Hermite normal form of M, upper-triangular integer matrix
    L: array, (m, m)
        unimodular matrix s.t. H = np.dot(L, M)
    """
    zmh = ZmoduleHomomorphism.with_standard_basis(M)
    return zmh.hermite_normal_form()


def column_style_hermite_normal_form(M):
    """
    calculate column-style Hermite normal form
    H = MR

    Parameters
    ----------
    M: array, (m, n)

    Returns
    -------
    H: array, (m, n)
        Hermite normal form of M, lower-triangular integer matrix
    R: array, (n, n)
        unimodular matrix s.t. H = np.dot(M, R)
    """
    zmh = ZmoduleHomomorphism.with_standard_basis(M.T)
    H_T, R_T = zmh.hermite_normal_form()
    H = H_T.T
    R = R_T.T
    return H, R



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Contains functions for working with findsym

Includes functions for running findsym, reading findsym input, and extracting
data from findsym output.
"""
from __future__ import print_function
from __future__ import division

from builtins import str
from builtins import range
from past.utils import old_div
import re
import math
import sys
import os
import subprocess

from fractions import Fraction
import sympy

def get_newfilename(path,filename):
    if not os.path.exists(path+'/'+filename):
        return filename
    else:
        new_filename = filename + '1'
        i = 1
        while os.path.exists(path+'/'+new_filename):
            i += 1
            new_filename = filename + '{}'.format(i)
    return new_filename

def read_fs_inp(inp,clean=True):
    with open(inp,'r') as f:

        fin = f.readlines()
        
        #fin_c cleans definition of axes from the findsym input
        #otherwise findsym crashes
        fin_c = []
        found = False
        i = 0
        for i in range(len(fin)):
            if 'axes:' in fin[i]:
                found = True
            if not found:
                fin_c.append(fin[i])

    if clean:
        return fin_c
    else:
        return fin

def run_fs_fin(fin):
    dirname, filename = os.path.split(os.path.abspath(__file__))
    my_env = os.environ.copy()
    isodata_path = dirname + '/findsym/'

    #findsym fails if the ISODATA variable is too long.
    #This is a simple fix, which tries to create a symbolic link in the current working directory
    try:
        cwd = os.getcwd()
        findsym_name = get_newfilename(cwd,'findsym')
        os.symlink(isodata_path,cwd+'/'+findsym_name)
        my_env["ISODATA"] = findsym_name+'/'
        link_created = True
    except:
        print('Cannot create a subdirectory in a current working directory. \
            Findym may fail if the path to findsym is too long.')
        my_env["ISODATA"] = isodata_path
        link_created = False

    try:  
        fs = subprocess.Popen([dirname+'/findsym/findsym'],stdin=subprocess.PIPE,stdout=subprocess.PIPE,env=my_env)
        out = fs.communicate(input=''.join(fin).encode())[0]
        lines = out.decode().split('\n')
        if link_created:
            try:
                os.unlink(findsym_name)
            except:
                print('Couldn not remove symlink {}'.format(cwd+'/'+findsym_name))
    except:
        if link_created:
            try:
                os.unlink(findsym_name)
            except:
                print('Couldn not remove symlink {}'.format(cwd+'/'+findsym_name))
        raise Exception('Error in findsym input') 

    return lines

def make_fsinp_nonmag(fin_c):
    start = False
    fin_cnm = []
    for i in range(len(fin_c)):
        if 'magnetic' in fin_c[i]:
            start = True
            fin_cnm.append(fin_c[i])
            continue
        if start:
            regex = r'([0-9\.\-eE]+ +[0-9\.\-eE]+ +[0-9\.\-eE]+) +[0-9\.\-eE]+ +[0-9\.\-eE]+ +[0-9\.\-eE]+'
            if len(re.findall(regex,fin_c[i])) == 0:
                print('here', fin_c[i])
                raise Exception('Problem when substituting zero moments')
            fin_cnm.append(re.sub(regex,r'\1 0 0 0',fin_c[i],count=1))
        else:
            fin_cnm.append(fin_c[i])
    return fin_cnm

def run_fs(inp):
    fin_c = read_fs_inp(inp)
    return run_fs_fin(fin_c)

def run_fs_nonmag(inp):
    
    fin_c = read_fs_inp(inp)
    #replaces the magnetic moments by 0
    fin_cnm = make_fsinp_nonmag(fin_c)
    
    return run_fs_fin(fin_cnm)

#tests if two vectors are equal with precision prec
def equal_vectors(vec1,vec2,prec):
     if len(vec1) != len(vec2):
         sys.exit('Comparing two vectors of unequal size:(')
     a = True
     for i in range(len(vec1)):
         if abs(vec1[i] - vec2[i]) > prec:
             a = False
     return a

#takes  a position (which includes a magnetic moment,) and symmetry and trasforms the position under the symmetry operation
#it transforms the magnetic moment too
#transformation of the magnetic moment is in principle not neccessary
def transform_position(pos,sym,prec):

     #transforms the position
     transd = []
     for i in range(3): #loop over components of position
         op = re.sub('-','+-',sym[1][i])
         transd.append(0)
         op = re.split('\+',op) #op contains a list of all operations we have to do with component
         op = [_f for _f in op if _f] #remove empty strings from the list  
         for j in range(len(op)):
             if re.match('x',op[j]):
                 transd[i] = transd[i] + float(pos[0])
             if re.match('-x',op[j]):
                 transd[i] = transd[i] - float(pos[0])
             if re.match('y',op[j]):
                 transd[i] = transd[i] + float(pos[1])
             if re.match('-y',op[j]):
                 transd[i] = transd[i] - float(pos[1])
             if re.match('z',op[j]):
                 transd[i] = transd[i] + float(pos[2])
             if re.match('-z',op[j]):
                 transd[i] = transd[i] - float(pos[2])
             if re.match('-?[0-9]*[./]?[0-9]+',op[j]):
                 if re.match('-?[0-9.]+/[0-9.]+',op[j]):
                     op_s = op[j].split('/')
                     op[j] = old_div(float(op_s[0]), float(op_s[1]))
                 transd[i] = transd[i] + float(op[j])

     #we want all positions to be in the first unit cell, ie they must lie between 0 and 1
     #if they are not we tranform it there
     for i in range(3):
         transd[i] = transd[i] - math.floor(transd[i])
         #due to rounding error it may happen that the vector still lies outside of the unit cell
         #we have to take this into account
         if math.floor(transd[i]+prec) != 0: #if this occurs it means trands[i] lies within prec to 1 so we can se it to 0
             transd[i] = 0

     #transforms the momentum
     transd_m = []
     for i in range(3): #loop over components of magnetic moment
         op = re.sub('-','+-',sym[2][i])
         transd_m.append(0)
         op = re.split('\+',op) #op contains a list of all operations we have to do with component
         op = [_f for _f in op if _f] #remove empty strings from the list  
         for j in range(len(op)):
             if re.match('mx',op[j]):
                 transd_m[i] = transd_m[i] + float(pos[3])
             if re.match('-mx',op[j]):
                 transd_m[i] = transd_m[i] - float(pos[3])
             if re.match('my',op[j]):
                 transd_m[i] = transd_m[i] + float(pos[4])
             if re.match('-my',op[j]):
                 transd_m[i] = transd_m[i] - float(pos[4])
             if re.match('mz',op[j]):
                 transd_m[i] = transd_m[i] + float(pos[5])
             if re.match('-mz',op[j]):
                 transd_m[i] = transd_m[i] - float(pos[5])
             if re.match('-?[0-9]*[./]?[0-9]+',op[j]):
                 if re.match('-?[0-9.]+/[0-9.]+',op[j]):
                     op_s = op[j].split('/')
                     op[j] = old_div(float(op_s[0]), float(op_s[1]))
                 transd_m[i] = transd_m[i] + float(op[j])

     return list(transd+transd_m)


def r_basis(lines):
#read basis transformation
     for i in range(len(lines)):
         if "Vectors a,b,c:" in lines[i]:
             pos_vectors = i

     vec1 = lines[pos_vectors + 1]
     vec2 = lines[pos_vectors + 2]
     vec3 = lines[pos_vectors + 3]

     vec_a = [float(x) for x in vec1.split()]
     vec_b = [float(x) for x in vec2.split()]
     vec_c = [float(x) for x in vec3.split()]

     return [vec_a,vec_b,vec_c]


def r_origin(lines):
     out = False
     #finds the origin
     for i in range(len(lines)):
         if "Origin at" in lines[i]:
             out = lines[i].split()[2:5]
     for i in range(3):
         out[i] = float(out[i])
     #if we don't find the origin, we assum it's at 0,0,0    
     if not out:
         out = [0,0,0]
     return out
                 


def  r_pos(lines, fix_m=[]):

    #read atomic positions
    #format: list containing six floats for each position first three are the positions the other three magnetic moment
    #fix_m: m is not given in basis of a,b,c, but in basis of unit vectrs along these axis:
        #m = m1*a/|a|+m2*b/|b|+m3*c/|c|
    #since it is more interesting to have m in terms of a,b,c, this can be changed
    #needs magnitudes of a,b,c as input: fix_m=[a,b,c]
    
    for i in range(len(lines)):
        if "Atomic positions and magnetic moments in terms of a,b,c:" in lines[i]:
            pos_pos = i

    cont = 1
    positions = []
    i = 0
    while cont == 1:
        i = i+1
        if re.match('\A +[0-9]+ ',lines[pos_pos + i]):
            positions.append(lines[pos_pos + i])
        elif re.match('-------',lines[pos_pos + i]):
            cont = 0

    for i in range(len(positions)):
        positions[i] = positions[i].split()
        at_number = int(positions[i].pop(0))
        for l in range(len(positions[i])):
            positions[i][l] = float(positions[i][l])
        positions[i].append(at_number)

    if fix_m:
        for i in range(len(positions)):
            for l in range(3):
                positions[i][l+3] = old_div(positions[i][l+3],fix_m[l])

    return positions


def r_sym(lines,debug=False,syms_only=False,num_prec=None):
#read symmetries 
#format: number of symmetry, space transformation, time transformation, switch controlling
#if there is time-reversal, list of tuples which show to which position each position transforms
#under this symmetry
     if num_prec is None:
         #tolerance = r_tolerance(lines)
         num_prec = 1.1e-5
     debug = False
     for i in range(len(lines)):
         if "_space_group_symop.magn_operation_mxmymz" in lines[i]:
             pos_syms = i

     cont = 1
     syms = []
     i = 0
     while cont == 1:
         i = i+1
         if re.match('\A[0-9]+ ',lines[pos_syms + i]):
             syms.append(lines[pos_syms + i])
         else:
             cont = 0

     for i in range(len(syms)):
         syms[i] = syms[i].split()
         syms[i][1] = syms[i][1].split(',')
         syms[i][2] = syms[i][2].split(',')
         syms[i].append(syms[i][1][3])
         syms[i][1] = syms[i][1][0:3]

     if debug:
         print('found symmetries:')
         for sym in syms:
             print(sym)

     if not syms_only:
         #read space group number
         for i in range(len(lines)):
             if "Space Group " in lines[i]:
                 group = re.findall('[0-9]+\.[0-9]+',lines[i])

         positions = r_pos(lines)
         check_pos_prec(positions,num_prec)

         if debug:
             print('group')
             print(group)
             print('positions')
             for pos in positions:
                 print(pos)
         #magnetic group tables
         #needs a file 'tables_wyckoff.txt' which are magnetic tables with all unneccessary information deleted
         dirname, filename = os.path.split(os.path.abspath(__file__))
         tables_loc = str(dirname)+'/findsym/tables_wyckoff.txt'
         tables=open(tables_loc)

         #read shifts of wickoff positions
         #this is neccessary to obtain all positions
         #in body centered structure for example, the output from findsym only contains half of the positions,
         #but symmetries also consider those that trasform one atom into another one shifted by 1/2 1/2 1/2
         found = False
         end = False
         for line in tables:
             rx = r"BNS: +" + re.escape(group[0])
             if re.match(rx,line):
                 found = True
                 if debug:
                     print(line)
             if found and not end:
                 if 'Wyckoff positions' in line:
                     end = True
                     shifts = re.findall('\([0-9./]+,[0-9./]+,[0-9./]+\)\+',line)
                     if debug:
                         print(line)

         if debug:
             print('shifts')
             print(shifts)
         if shifts:
             for l in range(len(shifts)):
                 shift_sep = re.findall('[0-9./]+',shifts[l])
                 for i in range(len(shift_sep)):
                     if re.match('[0-9]+/[0-9]+',shift_sep[i]):
                         t = shift_sep[i].split('/')
                         shift_sep[i] = old_div(float(t[0]), float(t[1]))
                     else:
                         shift_sep[i] = float(shift_sep[i])
                 shifts[l] = shift_sep

             if shifts[0] != [0,0,0]:
                 sys.exit('First shift is not zero')
             else:
                 shifts.pop(0)

                 if debug:
                     print('found wyckoff shifts:')
                     for shift in shifts:
                         print(shift)
         

         if debug:
             print('finding info about sublattice transformations')
         #this adds the information about transformation of sublattices 
         for l in range(len(syms)):
             if debug:
                 print('')
                 print('taking symmetry', syms[l])
             sym_trans = []
             for i in range(len(positions)):
                 trans = transform_position(positions[i],syms[l],num_prec)
                 if debug:
                     print('taking position:', positions[i])
                     print('transformed to:', trans)
                 found_atoms = []
                 for j in range(len(positions)):
                     if equal_vectors(trans[0:3],positions[j][0:3],num_prec):
                         found_atoms.append(positions[j][6])
                         #sym_trans.append((positions[i][6],positions[j][6]))
                         if debug:
                             print('transformed atom identified as atom ', positions[j][6], ' with position ', positions[j])
                     else:
                         for k in range(len(shifts)):
                             sym_temp =['',['x+'+str(shifts[k][0]),'y+'+str(shifts[k][1]),'z+'+str(shifts[k][2])],['mx','my','mz']]
                             pos_shift = transform_position(positions[j][0:6],sym_temp,num_prec)
                             if equal_vectors(trans[0:3],pos_shift[0:3],num_prec):
                                 #sym_trans.append((positions[i][6],positions[j][6]))
                                 found_atoms.append(positions[j][6])
                                 if debug:
                                     print('transformed atom identified as atom ', positions[j][6], ' with position ', positions[j])
                 if len(found_atoms) != 1:
                    print(syms[l])
                    print(found_atoms)
                    raise Exception('Problem with identifying transformation of atom {0}. '
                            'Try changing --pos-prec.'.format(positions[i][6]))
                 else:
                    sym_trans.append((positions[i][6],found_atoms[0]))
                        

             if len(sym_trans) != len(positions):
                 print(syms[l])
                 print(sym_trans)
                 raise Exception('Wrong number of transformed atoms. Try modifying --pos-prec parameter.')
             syms[l].append(list(sym_trans))

     return syms


def r_abc(lines):

     for i in range(len(lines)):
         if "Values of a,b,c,alpha,beta,gamma" in lines[i]:
             pos_abc = i

     abc = lines[pos_abc+1].split()

     return abc

def r_mag_fin(fin):
    start = False
    mags = []
    for i in range(len(fin)):
        if start:
            n += 1
            if n > n_at:
                break
            mag = sympy.zeros(1,3)  
            for j in range(3):
                mag[j] = sympy.sympify(fin[i].split()[j+3])
            mags.append(mag)
        elif (not start) and 'magnetic' in fin[i]:
            start = True
            n_at = int(fin[i-2])
            n = 0
    return mags

def r_tolerance(lines):
    for line in lines:
        if 'Tolerance:' in line:
            num_prec = float(line.split(':')[1])
    if abs(num_prec) < 1e-14:
        num_prec = 1e-6
    return num_prec

def check_pos_prec(positions,num_prec):
    for i,pos in enumerate(positions):
        for j,pos2 in enumerate(positions):
            if i != j:
                if equal_vectors(pos[0:3],pos2[0:3],num_prec):
                    raise Exception(
                            'Atoms {0} and {1} too close together try lower --pos-prec'.format(pos[6],pos2[6]))





import json
from symmetr import parse, sym_res
from symmetr.tensors import Tensor

def test_jE():
    opt = parse('res j E -g P-43m --ignore-same-op-sym')
    X = sym_res(opt)
    with open('groups/jE_-43m.out') as f:
        res = json.load(f)
    Xr0 = Tensor.load(res['X0'])
    Xr1 = Tensor.load(res['X1'])
    X_zero = Tensor(0,3,2)

    assert X[0] - Xr0 == X_zero
    assert X[1] - Xr1 == X_zero

def test_sE_exp():

    opt = parse('res s E -g P4mm --exp 1 ')
    X = sym_res(opt)
    with open('groups/sE_4mm_exp1.out') as f:
        res = json.load(f)
    Xr = Tensor.load(res['X'])
    X_zero = Tensor(0,3,2)

    assert X- Xr == X_zero




import setuptools

with open("README_pypi.md", "r") as fh:
    long_description = fh.read()

def get_version():
    with open("symmetr/version.py") as f:
        lines = f.readlines()
    for line in lines:
        if 'version' in line:
            version = line.split('=')[1].strip().lstrip('\'').rstrip('\'')
            return version

setuptools.setup(
    name="symmetr",
    version=get_version(),
    author="Jakub Zelezny",
    author_email="jakub.zelezny@gmail.com",
    description="Package for determining symmetry properties of crystals.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/zeleznyj/linear-response-symmetry",
    packages=setuptools.find_packages(),
    package_data={'symmetr': ['findsym/*']},
    data_files=[('',['README_pypi.md'])],
    scripts = ['exec/symmetr'],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        'future',
        'six',
        'numpy',
        'numpy; python_version>="3"',
        'numpy==1.16.4; python_version<"3"',
        'sympy',
        'prettytable',
        'scipy',
        'lxml',
        'pandas',
        'beautifulsoup4',
        'pymatgen'
    ]
)



from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
from builtins import object
import sympy
import re

class Symmetry(object):

    def __init__(self,R,has_T,Rs=None,permutations=None):
        self.R = R
        self.has_T = has_T
        if Rs is None:
            factor = 1
            if has_T:
                factor *= -1
            if self.R.det() == -1:
                factor *= -1
            self.Rs = factor * self.R
        else:
            if abs(abs(Rs.det())-1) > 1e-10:
                raise Warning('Rs must have determinant +1 or -1')
            self.Rs = Rs
        self.permutations = permutations
        self.custom_Rs = {}

    def get_R(self,op=None):
        R_op = None
        if op is None or op == 'x':
            R_op = self.R
        elif op == 's':
            R_op = self.Rs
        elif op == 'v':
            if not self.has_T:
                R_op = self.R
            else:
                R_op = -self.R
        elif op == 't':
            if not self.has_T:
                R_op = self.R.det() * self.R
            else:
                R_op = -self.R.det() * self.R
        elif op in self.custom_Rs:
            R_op = self.custom_Rs[op]

        if R_op is None:
            print(self.custom_Rs)
            print(op)
            raise Exception('wrong op')
        else:
            return R_op

    def convert(self,T,in_place=False):
        """
        Converts the symmetry operation to a different coordinate system.

        Args:
            T (matrix): Coordinate transformation matrix. If it is set, the symmetry operation will be
                        transformed by this matrix.
                        Symmetry operations are given in basis A. T transforms from A to B, ie Tx_A = x_B.

        Doesn't return anything, the symmetry itself is modified.
        """ 

        R_T = T * self.R * T.inv()
        Rs_T = T * self.Rs * T.inv()
        custom_Rs_T = {}
        for name in self.custom_Rs:
            custom_Rs_T[name] = T * self.custom_Rs[name] * T.inv()

        if in_place:
            self.R = R_T
            self.Rs = Rs_T
            for name in self.custom_Rs:
                self.custom_Rs[name] = self.custom_Rs_T[name]
        else:
            out = Symmetry(R=R_T,has_T=self.has_T,Rs=Rs_T,permutations=self.permutations)
            out.custom_Rs = custom_Rs_T
            return out

    def def_custom_R(self,name,R_c):
        self.custom_Rs[name] = R_c

    def copy(self):
        return Symmetry(self.R,self.has_T,self.Rs,self.permutations)

    def __str__(self):
        out = 'R: '
        out += self.R.__str__()  + '\n'
        out += 'Rs: '
        out += self.Rs.__str__() + '\n'
        out += 'has_T: ' + str(self.has_T) + '\n'
        out += 'permutations: ' + str(self.permutations)

        return out

    def __eq__(self,other):
        if self.R != other.R:
            return False
        if self.has_T != other.has_T:
            return False
        if self.Rs != other.Rs:
            return False
        if self.permutations is None or other.permutations is None:
            if not (self.permutations is None and other.permutations is None):
                return False
            if self.permutations != other.permutations:
                return False
        if self.custom_Rs != other.custom_Rs:
            return False
        return True

    def eq_numeric(self,other,prec=1e-4):
        if sympy.N(self.R - other.R).norm() > prec:
            return False
        if self.has_T != other.has_T:
            return False
        if sympy.N(self.Rs - other.Rs).norm() > prec:
                return False
        if self.permutations is None or other.permutations is None:
            if not (self.permutations is None and other.permutations is None):
                return False
            if self.permutations != other.permutations:
                return False
        for op in self.custom_Rs:
            if op not in other.custom_Rs:
                return False
            else:
                if sympy.N(self.custom_Rs[op] - other.custom_Rs[op]).norm() > prec:
                    return False
        return True

    def __mul__(self,other):
        R = self.R * other.R
        Rs = self.Rs * other.Rs
        if not self.has_T and not other.has_T:
            has_T = False
        elif self.has_T and other.has_T:
            has_T = False
        else:
            has_T = True
        if self.permutations is None and other.permutation is None:
            permutations = None
        else:
            if self.permutations is None or other.permutations is None:
                raise Exception('Cannot multiply symmetry operations when only one permutation is defined')
            permutations = {}
            for i in other.permutations:
                permutations[i] = self.permutations[other.permutations[i]]
        if self.custom_Rs.keys() != other.custom_Rs.keys():
            raise Exception('For multiplication the custom_Rs must be the same')
        custom_Rs = {}
        for name in custom_Rs:
            custom_Rs[name] = self.custom_Rs[name]*other.custom_Rs[name]
        out = Symmetry(R,has_T,Rs,permutations)
        out.custom_Rs = custom_Rs
        return out

    def pprint(self):
        sympy.pprint(self.R)
        sympy.pprint(self.Rs)
        for custom_R in self.custom_Rs:
            print(custom_R)
            sympy.pprint(self.custom_Rs[custom_R])
        print('has T:', self.has_T)
        if self.permutations is not None:
            print('permutations: ', self.permutations)

    def inv(self):
        Ri = self.R.inv()
        Rsi = self.Rs.inv()
        has_Ti = self.has_T
        if self.permutations is not None:
            permuationsi = {}
            for p in self.permutations:
                permuationsi[self.permutations[p]] = p
        else:
            permutationsi = None
        out = Symmetry(Ri,has_Ti,Rsi,permuationsi)
        for Rc in self.custom_Rs:
            out.custom_Rs[Rc] = self.custom_Rs[Rc].inv()

        return out

def findsym2sym(sym_findsym):
    R = sym2R(sym_findsym)
    Rs = sym2Rs(sym_findsym)
    if sym_findsym[3] == '-1':
        has_T = True
    elif sym_findsym[3] == '+1':
        has_T = False
    else:
        print(sym_findsym[3])
        raise Exception('Wrong findsym format')
    if len(sym_findsym) > 4:
        permutations = {}
        for (i,j) in sym_findsym[4]:
            permutations[i] = j
    else:
        permutations = None
    return Symmetry(R=R,Rs=Rs,has_T=has_T,permutations=permutations)

def matsym2sym(sym):
    R = sym[0]
    Rs = sym[2]
    if sym[3] == '-1':
        has_T = True
    elif sym[3] == '+1':
        has_T = False
    else:
        print(sym[3])
        raise Exception('Wrong findsym format')
    permutations = {}
    for (i,j) in sym[4]:
        permutations[i] = j
    return Symmetry(R=R,Rs=Rs,has_T=has_T,permutations=permutations)

def sym2R(sym):

    R = sympy.zeros(3)
    for i in range(3):
        trans = convert_op(sym,['x',i])
        for t in trans:
            for l in range(3):
                if t[0] == l:
                    R[i,l] = t[1]

    return R

def sym2Rs(sym):

    sym_s = sympy.zeros(3)
    for i in range(3):
        trans = convert_op(sym,['s',i])
        for t in trans:
            for l in range(3):
                if t[0] == l:
                    sym_s[i,l] = t[1]

    return sym_s

def sym2trans(sym):
    out = []
    for i in range(3):
        out.append(sym[1][i])
        for s in ['x','y','z']:
            out[i] = out[i].replace(s,'0')
        out[i] = sympy.sympify(out[i])
    return sympy.Matrix(out)

def create_T():

    R = sympy.diag(1,1,1)
    return Symmetry(R=R,has_T=True)

def create_I():
    R = sympy.diag(1,1,1)
    return Symmetry(R=R,has_T=False)

def create_P():
    """
    Note this doesn't contain any permutations, so it's not always correct!
    """
    R = -sympy.diag(1,1,1)
    return Symmetry(R=R,has_T=False)

def convert_op(sym,op_type):
    """
    Transforms operator component by a symmetry operation.

    Args:
        sym: The symmetry operation.
        op_type: Determines the operator type and operator component to be transformed.
            [operator type,component index(0,1 or 2)]

    Returns:
        out: A list of tuples. For example [(0,1),(1,-1)].
            First component means operator index. Second component means sign. The tuples are to be summed up.
            The example means: op_x-op_y
    """

    #velocity operator
    if op_type[0] == 'v':

        s = sym[1][op_type[1]]

        op = re.sub('-','+-',s)
        op = re.split('\+',op)
        op = [_f for _f in op if _f] #remove empty strings from the list  
        out = []
        for j in range(len(op)):
            match = False
            if re.match('^x',op[j]):
                t = (0,1)
                match = True
            if re.match('^-x',op[j]):
                t = (0,-1)
                match = True
            if re.match('^y',op[j]):
                t = (1,1)
                match = True
            if re.match('^-y',op[j]):
                t = (1,-1)
                match = True
            if re.match('^z',op[j]):
                t = (2,1)
                match = True
            if re.match('^-z',op[j]):
                t = (2,-1)
                match = True
            #if there is a time-reversal, v has a minus compared to space transformation
            if match:
                if sym[3] == '-1':
                    t = (t[0],-1*t[1])
                out.append(t)

        return out

    #spin operator
    if op_type[0] == 's':

        s = sym[2][op_type[1]]

        op = re.sub('-','+-',s)
        op = re.split('\+',op)
        op = [_f for _f in op if _f] #remove empty strings from the list  
        out = []
        for j in range(len(op)):
            if re.match('^mx',op[j]):
                t = (0,1)
                out.append(t)
            if re.match('^-mx',op[j]):
                t = (0,-1)
                out.append(t)
            if re.match('^my',op[j]):
                t = (1,1)
                out.append(t)
            if re.match('^-my',op[j]):
                t = (1,-1)
                out.append(t)
            if re.match('^mz',op[j]):
                t = (2,1)
                out.append(t)
            if re.match('^-mz',op[j]):
                t = (2,-1)
                out.append(t)

        return out

    #torque operator
    if op_type[0] == 't':

        s = sym[2][op_type[1]]

        op = re.sub('-','+-',s)
        op = re.split('\+',op)
        op = [_f for _f in op if _f] #remove empty strings from the list  
        out = []
        for j in range(len(op)):
            match = False
            if re.match('^mx',op[j]):
                t = (0,1)
                out.append(t)
                match = True
            if re.match('^-mx',op[j]):
                t = (0,-1)
                out.append(t)
                match = True
            if re.match('^my',op[j]):
                t = (1,1)
                out.append(t)
                match = True
            if re.match('^-my',op[j]):
                t = (1,-1)
                out.append(t)
                match = True
            if re.match('^mz',op[j]):
                t = (2,1)
                out.append(t)
                match = True
            if re.match('^-mz',op[j]):
                t = (2,-1)
                out.append(t)
                match = True
            if match:
                if sym[3] == '-1':
                    t = (t[0],-1*t[1])
                out.append(t)

        return out
    
    #transformation of a position operator
    if op_type[0] == 'x':

        s = sym[1][op_type[1]]

        op = re.sub('-','+-',s)
        op = re.split('\+',op)
        op = [_f for _f in op if _f] #remove empty strings from the list  
        out = []
        for j in range(len(op)):
            match = False
            if re.match('^x',op[j]):
                t = (0,1)
                match = True
            if re.match('^-x',op[j]):
                t = (0,-1)
                match = True
            if re.match('^y',op[j]):
                t = (1,1)
                match = True
            if re.match('^-y',op[j]):
                t = (1,-1)
                match = True
            if re.match('^z',op[j]):
                t = (2,1)
                match = True
            if re.match('^-z',op[j]):
                t = (2,-1)
                match = True
            if match:
                out.append(t)

        return out
    
    #finds a translational component of a transformation
    if op_type[0] == 'translation':
        
        s = sym[1][op_type[1]]

        op = re.sub('-','+-',s)
        op = re.split('\+',op)
        op = [_f for _f in op if _f] #remove empty strings from the list  
        out = []
        for j in range(len(op)):
            match = False
            if re.match('-?[0-9]*[./]?[0-9]+',op[j]):
                if re.match('-?[0-9.]+/[0-9.]+',op[j]):
                    op_s = op[j].split('/')
                    op[j] = old_div(float(op_s[0]), float(op_s[1]))
                    match = True
            if match:
                out.append(op[j])

        return out

